@@ -1,4 +1,4 @@
-/*	$KAME: crypto_openssl.c,v 1.83 2003/11/13 19:51:43 sakane Exp $	*/
+/* $Id: crypto_openssl.c,v 1.40 2004/12/29 14:23:08 vanhu Exp $ */
 
 /*
  * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
@@ -40,27 +40,17 @@
 #include <string.h>
 
 /* get openssl/ssleay version number */
-#ifdef HAVE_OPENSSL_OPENSSLV_H
-# include <openssl/opensslv.h>
-#else
-# error no opensslv.h found.
+#include <openssl/opensslv.h>
+
+#if !defined(OPENSSL_VERSION_NUMBER) || (OPENSSL_VERSION_NUMBER < 0x0090602fL)
+#error OpenSSL version 0.9.6 or later required.
 #endif
 
-#ifndef OPENSSL_VERSION_NUMBER
-#error OPENSSL_VERSION_NUMBER is not defined. OpenSSL0.9.4 or later required.
-#endif
-
-#ifdef HAVE_OPENSSL_PEM_H
 #include <openssl/pem.h>
-#endif
-#ifdef HAVE_OPENSSL_EVP_H
 #include <openssl/evp.h>
-#endif
-#ifdef HAVE_OPENSSL_X509_H
 #include <openssl/x509.h>
 #include <openssl/x509v3.h>
 #include <openssl/x509_vfy.h>
-#endif
 #include <openssl/bn.h>
 #include <openssl/dh.h>
 #include <openssl/md5.h>
@@ -68,13 +58,10 @@
 #include <openssl/hmac.h>
 #include <openssl/des.h>
 #include <openssl/crypto.h>
-#ifdef HAVE_OPENSSL_IDEA_H
-#include <openssl/idea.h>
+#ifdef HAVE_OPENSSL_ENGINE_H
+#include <openssl/engine.h>
 #endif
 #include <openssl/blowfish.h>
-#ifdef HAVE_OPENSSL_RC5_H
-#include <openssl/rc5.h>
-#endif
 #include <openssl/cast.h>
 #include <openssl/err.h>
 #if defined(HAVE_OPENSSL_AES_H)
@@ -84,11 +71,13 @@
 #else
 #include "crypto/rijndael/rijndael-api-fst.h"
 #endif
+#ifdef WITH_SHA2
 #ifdef HAVE_OPENSSL_SHA2_H
 #include <openssl/sha2.h>
 #else
 #include "crypto/sha2/sha2.h"
 #endif
+#endif
 
 /* 0.9.7 stuff? */
 #if OPENSSL_VERSION_NUMBER < 0x0090700fL
@@ -97,6 +86,8 @@ typedef STACK_OF(GENERAL_NAME) GENERAL_NAMES;
 #define USE_NEW_DES_API
 #endif
 
+#define OpenSSL_BUG()	do { plog(LLV_ERROR, LOCATION, NULL, "OpenSSL function failed\n"); } while(0)
+
 #include "var.h"
 #include "misc.h"
 #include "vmbuf.h"
@@ -110,15 +101,12 @@ typedef STACK_OF(GENERAL_NAME) GENERAL_NAMES;
  * necessary for SSLeay/OpenSSL portability.  It sucks.
  */
 
-#ifdef HAVE_SIGNING_C
 static int cb_check_cert_local __P((int, X509_STORE_CTX *));
 static int cb_check_cert_remote __P((int, X509_STORE_CTX *));
 static X509 *mem2x509 __P((vchar_t *));
-#endif
 
 static caddr_t eay_hmac_init __P((vchar_t *, const EVP_MD *));
 
-#ifdef HAVE_SIGNING_C
 /* X509 Certificate */
 /*
  * convert the string of the subject name into DER
@@ -126,7 +114,7 @@ static caddr_t eay_hmac_init __P((vchar_t *, const EVP_MD *));
  */
 vchar_t *
 eay_str2asn1dn(str, len)
-	char *str;
+	const char *str;
 	int len;
 {
 	X509_NAME *name;
@@ -136,6 +124,9 @@ eay_str2asn1dn(str, len)
 	vchar_t *ret;
 	caddr_t p;
 
+	if (len == -1)
+		len = strlen(str);
+
 	buf = racoon_malloc(len + 1);
 	if (!buf) {
 		printf("failed to allocate buffer\n");
@@ -154,12 +145,22 @@ eay_str2asn1dn(str, len)
 			continue;
 		} else if (buf[i] == ',' || buf[i] == '/') {
 			buf[i] = '\0';
-#if 0
-			printf("[%s][%s]\n", field, value);
-#endif
+
+			plog(LLV_DEBUG, LOCATION, NULL, "DN: %s=%s\n",
+			     field, value);
+
+			if (!value) goto err;
 			if (!X509_NAME_add_entry_by_txt(name, field,
-					MBSTRING_ASC, value, -1, -1, 0))
+					(value[0] == '*' && value[1] == 0) ? 
+						V_ASN1_PRINTABLESTRING : MBSTRING_ASC,
+					value, -1, -1, 0)) {
+				plog(LLV_ERROR, LOCATION, NULL, 
+				     "Invalid DN field: %s=%s\n",
+				     field, value);
+				plog(LLV_ERROR, LOCATION, NULL, 
+				     "%s\n", eay_strerror());
 				goto err;
+			}
 			for (j = i + 1; j < len; j++) {
 				if (buf[j] != ' ')
 					break;
@@ -170,12 +171,22 @@ eay_str2asn1dn(str, len)
 		}
 	}
 	buf[len] = '\0';
-#if 0
-	printf("[%s][%s]\n", field, value);
-#endif
+
+	plog(LLV_DEBUG, LOCATION, NULL, "DN: %s=%s\n",
+	     field, value);
+
+	if (!value) goto err;
 	if (!X509_NAME_add_entry_by_txt(name, field,
-			MBSTRING_ASC, value, -1, -1, 0))
+			(value[0] == '*' && value[1] == 0) ? 
+				V_ASN1_PRINTABLESTRING : MBSTRING_ASC,
+			value, -1, -1, 0)) {
+		plog(LLV_ERROR, LOCATION, NULL, 
+		     "Invalid DN field: %s=%s\n",
+		     field, value);
+		plog(LLV_ERROR, LOCATION, NULL, 
+		     "%s\n", eay_strerror());
 		goto err;
+	}
 
 	i = i2d_X509_NAME(name, NULL);
 	if (!i)
@@ -198,6 +209,182 @@ eay_str2asn1dn(str, len)
 	return NULL;
 }
 
+/*
+ * convert the hex string of the subject name into DER
+ */
+vchar_t *
+eay_hex2asn1dn(const char *hex, int len)
+{
+	BIGNUM *bn = BN_new();
+	char *binbuf;
+	size_t binlen;
+	vchar_t *ret = NULL;
+	
+	if (len == -1)
+		len = strlen(hex);
+
+	if (BN_hex2bn(&bn, hex) != len) {
+		plog(LLV_ERROR, LOCATION, NULL, 
+		     "conversion of Hex-encoded ASN1 string to binary failed: %s\n",
+		     eay_strerror());
+		goto out;
+	}
+	
+	binlen = BN_num_bytes(bn);
+	ret = vmalloc(binlen);
+	if (!ret) {
+		printf("failed to allocate buffer\n");
+		return NULL;
+	}
+	binbuf = ret->v;
+
+	BN_bn2bin(bn, binbuf);
+
+out:
+	BN_free(bn);
+
+	return ret;
+}
+
+/*
+ * The following are derived from code in crypto/x509/x509_cmp.c
+ * in OpenSSL0.9.7c:
+ * X509_NAME_wildcmp() adds wildcard matching to the original
+ * X509_NAME_cmp(), nocase_cmp() and nocase_spacenorm_cmp() are as is.
+ */
+#include <ctype.h>
+/* Case insensitive string comparision */
+static int nocase_cmp(const ASN1_STRING *a, const ASN1_STRING *b)
+{
+	int i;
+
+	if (a->length != b->length)
+		return (a->length - b->length);
+
+	for (i=0; i<a->length; i++)
+	{
+		int ca, cb;
+
+		ca = tolower(a->data[i]);
+		cb = tolower(b->data[i]);
+
+		if (ca != cb)
+			return(ca-cb);
+	}
+	return 0;
+}
+
+/* Case insensitive string comparision with space normalization 
+ * Space normalization - ignore leading, trailing spaces, 
+ *       multiple spaces between characters are replaced by single space  
+ */
+static int nocase_spacenorm_cmp(const ASN1_STRING *a, const ASN1_STRING *b)
+{
+	unsigned char *pa = NULL, *pb = NULL;
+	int la, lb;
+	
+	la = a->length;
+	lb = b->length;
+	pa = a->data;
+	pb = b->data;
+
+	/* skip leading spaces */
+	while (la > 0 && isspace(*pa))
+	{
+		la--;
+		pa++;
+	}
+	while (lb > 0 && isspace(*pb))
+	{
+		lb--;
+		pb++;
+	}
+
+	/* skip trailing spaces */
+	while (la > 0 && isspace(pa[la-1]))
+		la--;
+	while (lb > 0 && isspace(pb[lb-1]))
+		lb--;
+
+	/* compare strings with space normalization */
+	while (la > 0 && lb > 0)
+	{
+		int ca, cb;
+
+		/* compare character */
+		ca = tolower(*pa);
+		cb = tolower(*pb);
+		if (ca != cb)
+			return (ca - cb);
+
+		pa++; pb++;
+		la--; lb--;
+
+		if (la <= 0 || lb <= 0)
+			break;
+
+		/* is white space next character ? */
+		if (isspace(*pa) && isspace(*pb))
+		{
+			/* skip remaining white spaces */
+			while (la > 0 && isspace(*pa))
+			{
+				la--;
+				pa++;
+			}
+			while (lb > 0 && isspace(*pb))
+			{
+				lb--;
+				pb++;
+			}
+		}
+	}
+	if (la > 0 || lb > 0)
+		return la - lb;
+
+	return 0;
+}
+
+static int X509_NAME_wildcmp(const X509_NAME *a, const X509_NAME *b)
+{
+    int i,j;
+    X509_NAME_ENTRY *na,*nb;
+
+    if (sk_X509_NAME_ENTRY_num(a->entries)
+	!= sk_X509_NAME_ENTRY_num(b->entries))
+	    return sk_X509_NAME_ENTRY_num(a->entries)
+	      -sk_X509_NAME_ENTRY_num(b->entries);
+    for (i=sk_X509_NAME_ENTRY_num(a->entries)-1; i>=0; i--)
+    {
+	    na=sk_X509_NAME_ENTRY_value(a->entries,i);
+	    nb=sk_X509_NAME_ENTRY_value(b->entries,i);
+	    j=OBJ_cmp(na->object,nb->object);
+	    if (j) return(j);
+	    if ((na->value->length == 1 && na->value->data[0] == '*')
+	     || (nb->value->length == 1 && nb->value->data[0] == '*'))
+		    continue;
+	    j=na->value->type-nb->value->type;
+	    if (j) return(j);
+	    if (na->value->type == V_ASN1_PRINTABLESTRING)
+		    j=nocase_spacenorm_cmp(na->value, nb->value);
+	    else if (na->value->type == V_ASN1_IA5STRING
+		    && OBJ_obj2nid(na->object) == NID_pkcs9_emailAddress)
+		    j=nocase_cmp(na->value, nb->value);
+	    else
+		    {
+		    j=na->value->length-nb->value->length;
+		    if (j) return(j);
+		    j=memcmp(na->value->data,nb->value->data,
+			    na->value->length);
+		    }
+	    if (j) return(j);
+	    j=na->set-nb->set;
+	    if (j) return(j);
+    }
+
+    return(0);
+}
+
 /*
  * compare two subjectNames.
  * OUT:        0: equal
@@ -219,7 +406,7 @@ eay_cmp_asn1dn(n1, n2)
 	if (!d2i_X509_NAME(&b, (void *)&p, n2->l))
 		goto end;
 
-	i = X509_NAME_cmp(a, b);
+	i = X509_NAME_wildcmp(a, b);
 
     end:
 	if (a)
@@ -233,28 +420,18 @@ eay_cmp_asn1dn(n1, n2)
  * this functions is derived from apps/verify.c in OpenSSL0.9.5
  */
 int
-eay_check_x509cert(cert, CApath, local)
+eay_check_x509cert(cert, CApath, CAfile, local)
 	vchar_t *cert;
 	char *CApath;
+	char *CAfile;
 	int local;
 {
 	X509_STORE *cert_ctx = NULL;
 	X509_LOOKUP *lookup = NULL;
 	X509 *x509 = NULL;
-#if OPENSSL_VERSION_NUMBER >= 0x00905100L
 	X509_STORE_CTX *csc;
-#else
-	X509_STORE_CTX csc;
-#endif
 	int error = -1;
 
-	/* XXX define only functions required. */
-#if OPENSSL_VERSION_NUMBER >= 0x00905100L
-	OpenSSL_add_all_algorithms();
-#else
-	SSLeay_add_all_algorithms();
-#endif
-
 	cert_ctx = X509_STORE_new();
 	if (cert_ctx == NULL)
 		goto end;
@@ -267,7 +444,9 @@ eay_check_x509cert(cert, CApath, local)
 	lookup = X509_STORE_add_lookup(cert_ctx, X509_LOOKUP_file());
 	if (lookup == NULL)
 		goto end;
-	X509_LOOKUP_load_file(lookup, NULL, X509_FILETYPE_DEFAULT); /* XXX */
+
+	X509_LOOKUP_load_file(lookup, CAfile, 
+	    (CAfile == NULL) ? X509_FILETYPE_DEFAULT : X509_FILETYPE_PEM);
 
 	lookup = X509_STORE_add_lookup(cert_ctx, X509_LOOKUP_hash_dir());
 	if (lookup == NULL)
@@ -284,7 +463,6 @@ eay_check_x509cert(cert, CApath, local)
 	if (x509 == NULL)
 		goto end;
 
-#if OPENSSL_VERSION_NUMBER >= 0x00905100L
 	csc = X509_STORE_CTX_new();
 	if (csc == NULL)
 		goto end;
@@ -295,11 +473,6 @@ eay_check_x509cert(cert, CApath, local)
 #endif
 	error = X509_verify_cert(csc);
 	X509_STORE_CTX_cleanup(csc);
-#else
-	X509_STORE_CTX_init(&csc, cert_ctx, x509, NULL);
-	error = X509_verify_cert(&csc);
-	X509_STORE_CTX_cleanup(&csc);
-#endif
 
 	/*
 	 * if x509_verify_cert() is successful then the value of error is
@@ -319,11 +492,8 @@ end:
 }
 
 /*
- * Callback function for verifing certificate.
- * Derived from cb() in openssl/apps/s_server.c
- *
- * This one is called for certificates obtained from 
- * 'peers_certfile' directive.
+ * callback function for verifing certificate.
+ * this function is derived from cb() in openssl/apps/s_server.c
  */
 static int
 cb_check_cert_local(ok, ctx)
@@ -346,10 +516,10 @@ cb_check_cert_local(ok, ctx)
 		switch (ctx->error) {
 		case X509_V_ERR_CERT_HAS_EXPIRED:
 		case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
-#if OPENSSL_VERSION_NUMBER >= 0x00905100L
+		case X509_V_ERR_INVALID_CA:
+		case X509_V_ERR_PATH_LENGTH_EXCEEDED:
 		case X509_V_ERR_INVALID_PURPOSE:
 		case X509_V_ERR_UNABLE_TO_GET_CRL:
-#endif
 			ok = 1;
 			log_tag = LLV_WARNING;
 			break;
@@ -369,8 +539,8 @@ cb_check_cert_local(ok, ctx)
 }
 
 /*
- * Similar to cb_check_cert_local() but this one is called 
- * for certificates obtained from the IKE payload.
+ * callback function for verifing remote certificates.
+ * this function is derived from cb() in openssl/apps/s_server.c
  */
 static int
 cb_check_cert_remote(ok, ctx)
@@ -385,7 +555,6 @@ cb_check_cert_remote(ok, ctx)
 				X509_get_subject_name(ctx->current_cert),
 				buf,
 				256);
-
 		switch (ctx->error) {
 		case X509_V_ERR_UNABLE_TO_GET_CRL:
 			ok = 1;
@@ -462,9 +631,9 @@ eay_get_x509subjectaltname(cert, altname, type, pos)
 	int pos;
 {
 	X509 *x509 = NULL;
-        GENERAL_NAMES *gens;
+	GENERAL_NAMES *gens = NULL;
 	GENERAL_NAME *gen;
-	int i, len;
+	int len;
 	int error = -1;
 
 	*altname = NULL;
@@ -474,41 +643,70 @@ eay_get_x509subjectaltname(cert, altname, type, pos)
 	if (x509 == NULL)
 		goto end;
 
-        gens = X509_get_ext_d2i(x509, NID_subject_alt_name, NULL, NULL);
+	gens = X509_get_ext_d2i(x509, NID_subject_alt_name, NULL, NULL);
 	if (gens == NULL)
 		goto end;
 
-        for(i = 0; i < sk_GENERAL_NAME_num(gens); i++) {
-		if (i + 1 != pos)
-			continue;
-		break;
-	}
-
 	/* there is no data at "pos" */
-        if (i == sk_GENERAL_NAME_num(gens))
+	if (pos > sk_GENERAL_NAME_num(gens))
 		goto end;
 
-	gen = sk_GENERAL_NAME_value(gens, i);
+	gen = sk_GENERAL_NAME_value(gens, pos - 1);
 
-	/* make sure if the data is terminated by '\0'. */
-	if (gen->d.ia5->data[gen->d.ia5->length] != '\0') {
-		plog(LLV_ERROR, LOCATION, NULL,
-			"data is not terminated by '\0'.");
-		hexdump(gen->d.ia5->data, gen->d.ia5->length + 1);
-		goto end;
+	/* read DNSName / Email */
+	if (gen->type == GEN_DNS	||
+		gen->type == GEN_EMAIL	||
+		gen->type == GEN_URI )
+	{
+		/* make sure if the data is terminated by '\0'. */
+		if (gen->d.ia5->data[gen->d.ia5->length] != '\0')
+		{
+			plog(LLV_ERROR, LOCATION, NULL,
+				 "data is not terminated by NUL.");
+			hexdump(gen->d.ia5->data, gen->d.ia5->length + 1);
+			goto end;
+		}
+		
+		len = gen->d.ia5->length + 1;
+		*altname = racoon_malloc(len);
+		if (!*altname)
+			goto end;
+		
+		strlcpy(*altname, gen->d.ia5->data, len);
+		*type = gen->type;
+		error = 0;
 	}
+	/* read IP address */
+	else if (gen->type == GEN_IPADD)
+	{
+		unsigned char p[5], *ip;
+		ip = p;
+		
+		/* only support IPv4 */
+		if (gen->d.ip->length != 4)
+			goto end;
+		
+		/* convert Octet String to String
+		 * XXX ???????
+		 */
+		/*i2d_ASN1_OCTET_STRING(gen->d.ip,&ip);*/
+		ip = gen->d.ip->data;
 
-	len = gen->d.ia5->length + 1;
-	*altname = racoon_malloc(len);
-	if (!*altname)
-		goto end;
+		/* XXX Magic, enough for an IPv4 address
+		 */
+		*altname = racoon_malloc(20);
+		if (!*altname)
+			goto end;
+		
+		sprintf(*altname, "%u.%u.%u.%u", ip[0], ip[1], ip[2], ip[3]);
+		*type = gen->type;
+		error = 0;
+	}
+	/* XXX other possible types ?
+	 * For now, error will be -1 if unsupported type
+	 */
 
-	strlcpy(*altname, gen->d.ia5->data, len);
-	*type = gen->type;
-
-	error = 0;
-
-   end:
+end:
 	if (error) {
 		if (*altname) {
 			racoon_free(*altname);
@@ -518,10 +716,14 @@ eay_get_x509subjectaltname(cert, altname, type, pos)
 	}
 	if (x509)
 		X509_free(x509);
+	if (gens)
+		/* free the whole stack. */
+		sk_GENERAL_NAME_pop_free(gens, GENERAL_NAME_free);
 
 	return error;
 }
 
+
 /*
  * decode a X509 certificate and make a readable text terminated '\n'.
  * return the buffer allocated, so must free it later.
@@ -633,11 +835,7 @@ eay_get_x509cert(path)
 	fp = fopen(path, "r");
 	if (fp == NULL)
 		return NULL;
-#if OPENSSL_VERSION_NUMBER >= 0x00904100L
 	x509 = PEM_read_X509(fp, NULL, NULL, NULL);
-#else
-	x509 = PEM_read_X509(fp, NULL, NULL);
-#endif
 	fclose (fp);
 
 	if (x509 == NULL)
@@ -653,30 +851,14 @@ eay_get_x509cert(path)
 	error = i2d_X509(x509, &bp);
 	X509_free(x509);
 
-	if (error == 0)
+	if (error == 0) {
+		vfree(cert);
 		return NULL;
+	}
 
 	return cert;
 }
 
-/*
- * sign a souce by X509 signature.
- * XXX: to be get hash type from my cert ?
- *	to be handled EVP_dss().
- */
-vchar_t *
-eay_get_x509sign(source, privkey, cert)
-	vchar_t *source;
-	vchar_t *privkey;
-	vchar_t *cert;
-{
-	vchar_t *sig = NULL;
-
-	sig = eay_rsa_sign(source, privkey);
-
-	return sig;
-}
-
 /*
  * check a X509 signature
  *	XXX: to be get hash type from my cert ?
@@ -693,6 +875,7 @@ eay_check_x509sign(source, sig, cert)
 	X509 *x509;
 	u_char *bp;
 	EVP_PKEY *evp;
+	int res;
 
 	bp = cert->v;
 
@@ -708,58 +891,25 @@ eay_check_x509sign(source, sig, cert)
 		return -1;
 	}
 
-	return eay_rsa_verify(source, sig, evp);
-}
-
-/*
- * check a signature by signed with PKCS7 certificate.
- *	XXX: to be get hash type from my cert ?
- *		to be handled EVP_dss().
- * OUT: return -1 when error.
- *	0
- */
-int
-eay_check_pkcs7sign(source, sig, cert)
-	vchar_t *source;
-	vchar_t *sig;
-	vchar_t *cert;
-{
-	X509 *x509;
-	EVP_MD_CTX md_ctx;
-	EVP_PKEY *evp;
-	int error;
-	BIO *bio = BIO_new(BIO_s_mem());
-	char *bp;
-
-	if (bio == NULL)
-		return -1;
-	error = BIO_write(bio, cert->v, cert->l);
-	if (error != cert->l)
-		return -1;
-
-	bp = cert->v;
-	x509 = PEM_read_bio_X509(bio, NULL, NULL, NULL);
-	BIO_free(bio);
-	if (x509 == NULL)
-		return -1;
-
-	evp = X509_get_pubkey(x509);
-	X509_free(x509);
-	if (evp == NULL)
-		return -1;
-
-	/* Verify the signature */
-	/* XXX: to be handled EVP_dss() */
-	EVP_VerifyInit(&md_ctx, EVP_sha1());
-	EVP_VerifyUpdate(&md_ctx, source->v, source->l);
-	error = EVP_VerifyFinal(&md_ctx, sig->v, sig->l, evp);
+	res = eay_rsa_verify(source, sig, evp->pkey.rsa);
 
 	EVP_PKEY_free(evp);
 
-	if (error != 1)
-		return -1;
+	return res;
+}
 
-	return 0;
+/*
+ * check RSA signature
+ * OUT: return -1 when error.
+ *	0 on success
+ */
+int
+eay_check_rsasign(source, sig, rsa)
+	vchar_t *source;
+	vchar_t *sig;
+	RSA *rsa;
+{
+	return eay_rsa_verify(source, sig, rsa);
 }
 
 /*
@@ -781,11 +931,8 @@ eay_get_pkcs1privkey(path)
 	if (fp == NULL)
 		return NULL;
 
-#if OPENSSL_VERSION_NUMBER >= 0x00904100L
 	evp = PEM_read_PrivateKey(fp, NULL, NULL, NULL);
-#else
-	evp = PEM_read_PrivateKey(fp, NULL, NULL);
-#endif
+
 	fclose (fp);
 
 	if (evp == NULL)
@@ -835,11 +982,8 @@ eay_get_pkcs1pubkey(path)
 	if (fp == NULL)
 		return NULL;
 
-#if OPENSSL_VERSION_NUMBER >= 0x00904100L
 	x509 = PEM_read_X509(fp, NULL, NULL, NULL);
-#else
-	x509 = PEM_read_X509(fp, NULL, NULL);
-#endif
+
 	fclose (fp);
 
 	if (x509 == NULL)
@@ -872,11 +1016,10 @@ end:
 
 	return pkey;
 }
-#endif
 
 vchar_t *
-eay_rsa_sign(src, privkey)
-	vchar_t *src, *privkey; 
+eay_get_x509sign(src, privkey)
+	vchar_t *src, *privkey;
 {
 	EVP_PKEY *evp;
 	u_char *bp = privkey->v;
@@ -889,17 +1032,36 @@ eay_rsa_sign(src, privkey)
 	if (evp == NULL)
 		return NULL;
 
-	/* XXX: to be handled EVP_dss() */
-	/* XXX: Where can I get such parameters ?  From my cert ? */
+	sig = eay_rsa_sign(src, evp->pkey.rsa);
 
-	len = RSA_size(evp->pkey.rsa);
+	EVP_PKEY_free(evp);
+
+	return sig;
+}
+
+vchar_t *
+eay_get_rsasign(src, rsa)
+	vchar_t *src;
+	RSA *rsa;
+{
+	return eay_rsa_sign(src, rsa);
+}
+
+vchar_t *
+eay_rsa_sign(vchar_t *src, RSA *rsa)
+{
+	int len;
+	vchar_t *sig = NULL;
+	int pad = RSA_PKCS1_PADDING;
+
+	len = RSA_size(rsa);
 
 	sig = vmalloc(len);
 	if (sig == NULL)
 		return NULL;
 
-	len = RSA_private_encrypt(src->l, src->v, sig->v, evp->pkey.rsa, pad);
-	EVP_PKEY_free(evp);
+	len = RSA_private_encrypt(src->l, src->v, sig->v, rsa, pad);
+
 	if (len == 0 || len != sig->l) {
 		vfree(sig);
 		sig = NULL;
@@ -909,28 +1071,25 @@ eay_rsa_sign(src, privkey)
 }
 
 int
-eay_rsa_verify(src, sig, evp)
+eay_rsa_verify(src, sig, rsa)
 	vchar_t *src, *sig;
-	EVP_PKEY *evp;
+	RSA *rsa;
 {
 	vchar_t *xbuf = NULL;
 	int pad = RSA_PKCS1_PADDING;
 	int len = 0;
 	int error;
 
-	len = RSA_size(evp->pkey.rsa);
+	len = RSA_size(rsa);
 	xbuf = vmalloc(len);
 	if (xbuf == NULL) {
 		plog(LLV_ERROR, LOCATION, NULL, "%s\n", eay_strerror());
-		EVP_PKEY_free(evp);
 		return -1;
 	}
 
-	len = RSA_public_decrypt(sig->l, sig->v, xbuf->v, evp->pkey.rsa, pad);
-	if (len == 0 || len != src->l)
-		plog(LLV_ERROR, LOCATION, NULL, "%s\n", eay_strerror());
-	EVP_PKEY_free(evp);
+	len = RSA_public_decrypt(sig->l, sig->v, xbuf->v, rsa, pad);
 	if (len == 0 || len != src->l) {
+		plog(LLV_ERROR, LOCATION, NULL, "%s\n", eay_strerror());
 		vfree(xbuf);
 		return -1;
 	}
@@ -954,11 +1113,7 @@ eay_strerror()
 	int len = 0, n;
 	unsigned long l;
 	char buf[200];
-#if OPENSSL_VERSION_NUMBER >= 0x00904100L
 	const char *file, *data;
-#else
-	char *file, *data;
-#endif
 	int line, flags;
 	unsigned long es;
 
@@ -979,10 +1134,55 @@ eay_strerror()
 	return ebuf;
 }
 
-void
-eay_init_error()
+vchar_t *
+evp_crypt(vchar_t *data, vchar_t *key, vchar_t *iv, const EVP_CIPHER *e, int enc)
 {
-	ERR_load_crypto_strings();
+	vchar_t *res;
+	EVP_CIPHER_CTX ctx;
+
+	if (!e)
+		return NULL;
+
+	if (data->l % EVP_CIPHER_block_size(e))
+		return NULL;
+
+	if ((res = vmalloc(data->l)) == NULL)
+		return NULL;
+
+	EVP_CIPHER_CTX_init(&ctx);
+
+	if (!EVP_CipherInit(&ctx, e, key->v, iv->v, enc)) {
+		OpenSSL_BUG();
+		vfree(res);
+		return NULL;
+	}
+	
+	if (!EVP_Cipher(&ctx, res->v, data->v, data->l)) {
+		OpenSSL_BUG();
+		vfree(res);
+		return NULL;
+	}
+
+	EVP_CIPHER_CTX_cleanup(&ctx);
+
+	return res;
+}
+
+int
+evp_weakkey(vchar_t *key, const EVP_CIPHER *e)
+{
+	return 0;
+}
+
+int
+evp_keylen(int len, const EVP_CIPHER *e)
+{
+	if (!e)
+		return -1;
+	if (len != 0 && len != EVP_CIPHER_key_length(e))
+		return -1;
+	
+	return EVP_CIPHER_key_length(e);
 }
 
 /*
@@ -992,73 +1192,14 @@ vchar_t *
 eay_des_encrypt(data, key, iv)
 	vchar_t *data, *key, *iv;
 {
-	vchar_t *res;
-#ifdef USE_NEW_DES_API
-	DES_key_schedule ks;
-#else
-	des_key_schedule ks;
-#endif
-
-#ifndef USE_NEW_DES_API
-	if (data->l % 8)
-		return NULL;
-#endif
-
-#ifdef USE_NEW_DES_API
-	if (DES_key_sched((void *)key->v, &ks) != 0)
-#else
-	if (des_key_sched((void *)key->v, ks) != 0)
-#endif
-		return NULL;
-
-	/* allocate buffer for result */
-	if ((res = vmalloc(data->l)) == NULL)
-		return NULL;
-
-	/* decryption data */
-#ifdef USE_NEW_DES_API
-	DES_cbc_encrypt((void *)data->v, (void *)res->v, data->l,
-	                &ks, (void *)iv->v, DES_ENCRYPT);
-#else
-	des_cbc_encrypt((void *)data->v, (void *)res->v, data->l,
-	                ks, (void *)iv->v, DES_ENCRYPT);
-#endif
-
-	return res;
+	return evp_crypt(data, key, iv, EVP_des_cbc(), 1);
 }
 
 vchar_t *
 eay_des_decrypt(data, key, iv)
 	vchar_t *data, *key, *iv;
 {
-	vchar_t *res;
-#ifdef USE_NEW_DES_API
-	DES_key_schedule ks;
-#else
-	des_key_schedule ks;
-#endif
-
-#ifdef USE_NEW_DES_API
-	if (DES_key_sched((void *)key->v, &ks) != 0)
-#else
-	if (des_key_sched((void *)key->v, ks) != 0)
-#endif
-		return NULL;
-
-	/* allocate buffer for result */
-	if ((res = vmalloc(data->l)) == NULL)
-		return NULL;
-
-	/* decryption data */
-#ifdef USE_NEW_DES_API
-	DES_cbc_encrypt((void *)data->v, (void *)res->v, data->l,
-	                &ks, (void *)iv->v, DES_DECRYPT);
-#else
-	des_cbc_encrypt((void *)data->v, (void *)res->v, data->l,
-	                ks, (void *)iv->v, DES_DECRYPT);
-#endif
-
-	return res;
+	return evp_crypt(data, key, iv, EVP_des_cbc(), 0);
 }
 
 int
@@ -1076,73 +1217,9 @@ int
 eay_des_keylen(len)
 	int len;
 {
-	if (len != 0 && len != 64)
-		return -1;
-	return 64;
+	return evp_keylen(len, EVP_des_cbc());
 }
 
-#ifdef HAVE_OPENSSL_IDEA_H
-/*
- * IDEA-CBC
- */
-vchar_t *
-eay_idea_encrypt(data, key, iv)
-	vchar_t *data, *key, *iv;
-{
-	vchar_t *res;
-	IDEA_KEY_SCHEDULE ks;
-
-	idea_set_encrypt_key(key->v, &ks);
-
-	/* allocate buffer for result */
-	if ((res = vmalloc(data->l)) == NULL)
-		return NULL;
-
-	/* decryption data */
-	idea_cbc_encrypt(data->v, res->v, data->l,
-	                &ks, iv->v, IDEA_ENCRYPT);
-
-	return res;
-}
-
-vchar_t *
-eay_idea_decrypt(data, key, iv)
-	vchar_t *data, *key, *iv;
-{
-	vchar_t *res;
-	IDEA_KEY_SCHEDULE ks, dks;
-
-	idea_set_encrypt_key(key->v, &ks);
-	idea_set_decrypt_key(&ks, &dks);
-
-	/* allocate buffer for result */
-	if ((res = vmalloc(data->l)) == NULL)
-		return NULL;
-
-	/* decryption data */
-	idea_cbc_encrypt(data->v, res->v, data->l,
-	                &dks, iv->v, IDEA_DECRYPT);
-
-	return res;
-}
-
-int
-eay_idea_weakkey(key)
-	vchar_t *key;
-{
-	return 0;	/* XXX */
-}
-
-int
-eay_idea_keylen(len)
-	int len;
-{
-	if (len != 0 && len != 128)
-		return -1;
-	return 128;
-}
-#endif
-
 /*
  * BLOWFISH-CBC
  */
@@ -1150,40 +1227,14 @@ vchar_t *
 eay_bf_encrypt(data, key, iv)
 	vchar_t *data, *key, *iv;
 {
-	vchar_t *res;
-	BF_KEY ks;
-
-	BF_set_key(&ks, key->l, key->v);
-
-	/* allocate buffer for result */
-	if ((res = vmalloc(data->l)) == NULL)
-		return NULL;
-
-	/* decryption data */
-	BF_cbc_encrypt(data->v, res->v, data->l,
-		&ks, iv->v, BF_ENCRYPT);
-
-	return res;
+	return evp_crypt(data, key, iv, EVP_bf_cbc(), 1);
 }
 
 vchar_t *
 eay_bf_decrypt(data, key, iv)
 	vchar_t *data, *key, *iv;
 {
-	vchar_t *res;
-	BF_KEY ks;
-
-	BF_set_key(&ks, key->l, key->v);
-
-	/* allocate buffer for result */
-	if ((res = vmalloc(data->l)) == NULL)
-		return NULL;
-
-	/* decryption data */
-	BF_cbc_encrypt(data->v, res->v, data->l,
-		&ks, iv->v, BF_DECRYPT);
-
-	return res;
+	return evp_crypt(data, key, iv, EVP_bf_cbc(), 0);
 }
 
 int
@@ -1204,209 +1255,21 @@ eay_bf_keylen(len)
 	return len;
 }
 
-#ifdef HAVE_OPENSSL_RC5_H
-/*
- * RC5-CBC
- */
-vchar_t *
-eay_rc5_encrypt(data, key, iv)
-	vchar_t *data, *key, *iv;
-{
-	vchar_t *res;
-	RC5_32_KEY ks;
-
-	/* in RFC 2451, there is information about the number of round. */
-	RC5_32_set_key(&ks, key->l, key->v, 16);
-
-	/* allocate buffer for result */
-	if ((res = vmalloc(data->l)) == NULL)
-		return NULL;
-
-	/* decryption data */
-	RC5_32_cbc_encrypt(data->v, res->v, data->l,
-		&ks, iv->v, RC5_ENCRYPT);
-
-	return res;
-}
-
-vchar_t *
-eay_rc5_decrypt(data, key, iv)
-	vchar_t *data, *key, *iv;
-{
-	vchar_t *res;
-	RC5_32_KEY ks;
-
-	/* in RFC 2451, there is information about the number of round. */
-	RC5_32_set_key(&ks, key->l, key->v, 16);
-
-	/* allocate buffer for result */
-	if ((res = vmalloc(data->l)) == NULL)
-		return NULL;
-
-	/* decryption data */
-	RC5_32_cbc_encrypt(data->v, res->v, data->l,
-		&ks, iv->v, RC5_DECRYPT);
-
-	return res;
-}
-
-int
-eay_rc5_weakkey(key)
-	vchar_t *key;
-{
-	return 0;	/* No known weak keys when used with 16 rounds. */
-
-}
-
-int
-eay_rc5_keylen(len)
-	int len;
-{
-	if (len == 0)
-		return 128;
-	if (len < 40 || len > 2040)
-		return -1;
-	return len;
-}
-#endif
-
 /*
  * 3DES-CBC
  */
-static vchar_t *
-eay_3des_expand_key (key, size)
-	vchar_t *key;
-	size_t size;
-{
-	vchar_t *newkey;
-	char *ptr, *end;
-	size_t idx;
-	
-	if (key->l <= 0)
-	  return NULL;
-
-	if (key->l == size) {
-	  newkey = vdup (key);
-	  return newkey;
-	}
-	
-	newkey = vmalloc (size);
-	if (! newkey)
-	  return NULL;
-
-	ptr = newkey->v;
-	end = newkey->v + newkey->l;
-	idx = 0;
-	while (ptr < end) {
-	  size_t amount;
-	  amount = key->l > (end - ptr) ? (end - ptr) : key->l;
-	  memcpy (ptr, key->v, amount);
-	  ptr += amount;
-	}
-
-	return newkey;
-}
-
 vchar_t *
 eay_3des_encrypt(data, key, iv)
 	vchar_t *data, *key, *iv;
 {
-	vchar_t *res, *expkey;
-
-#ifdef USE_NEW_DES_API
-	DES_key_schedule ks1, ks2, ks3;
-#else
-	des_key_schedule ks1, ks2, ks3;
-#endif
-
-	expkey = eay_3des_expand_key (key, 24);
-
-	if (expkey == NULL)
-	  return NULL;
-
-#ifdef USE_NEW_DES_API
-	if (DES_key_sched((void *)expkey->v, &ks1) != 0)
-		return NULL;
-	if (DES_key_sched((void *)(expkey->v + 8), &ks2) != 0)
-		return NULL;
-	if (DES_key_sched((void *)(expkey->v + 16), &ks3) != 0)
-		return NULL;
-#else
-	if (des_key_sched((void *)expkey->v, ks1) != 0)
-		return NULL;
-	if (des_key_sched((void *)(expkey->v + 8), ks2) != 0)
-		return NULL;
-	if (des_key_sched((void *)(expkey->v + 16), ks3) != 0)
-		return NULL;
-#endif
-
-	/* allocate buffer for result */
-	if ((res = vmalloc(data->l)) == NULL)
-		return NULL;
-
-	/* decryption data */
-#ifdef USE_NEW_DES_API
-	DES_ede3_cbc_encrypt((void *)data->v, (void *)res->v, data->l,
-	                &ks1, &ks2, &ks3, (void *)iv->v, DES_ENCRYPT);
-#else
-	des_ede3_cbc_encrypt((void *)data->v, (void *)res->v, data->l,
-	                ks1, ks2, ks3, (void *)iv->v, DES_ENCRYPT);
-#endif
-
-	vfree (expkey);
-
-	return res;
+	return evp_crypt(data, key, iv, EVP_des_ede3_cbc(), 1);
 }
 
 vchar_t *
 eay_3des_decrypt(data, key, iv)
 	vchar_t *data, *key, *iv;
 {
-	vchar_t *res, *expkey;
-
-#ifdef USE_NEW_DES_API
-	DES_key_schedule ks1, ks2, ks3;
-#else
-	des_key_schedule ks1, ks2, ks3;
-#endif
-
-	expkey = eay_3des_expand_key (key, 24);
-
-	if (expkey == NULL)
-	  return NULL;
-
-#ifdef USE_NEW_DES_API
-	if (DES_key_sched((void *)expkey->v, &ks1) != 0)
-		return NULL;
-	if (DES_key_sched((void *)(expkey->v + 8), &ks2) != 0)
-		return NULL;
-	if (DES_key_sched((void *)(expkey->v + 16), &ks3) != 0)
-		return NULL;
-#else
-	if (des_key_sched((void *)expkey->v, ks1) != 0)
-		return NULL;
-	if (des_key_sched((void *)(expkey->v + 8), ks2) != 0)
-		return NULL;
-	if (des_key_sched((void *)(expkey->v + 16), ks3) != 0)
-		return NULL;
-#endif
-
-	/* allocate buffer for result */
-	if ((res = vmalloc(data->l)) == NULL)
-		return NULL;
-
-	/* decryption data */
-#ifdef USE_NEW_DES_API
-	DES_ede3_cbc_encrypt((void *)data->v, (void *)res->v, data->l,
-	                &ks1, &ks2, &ks3, (void *)iv->v, DES_DECRYPT);
-#else
-	des_ede3_cbc_encrypt((void *)data->v, (void *)res->v, data->l,
-	                ks1, ks2, ks3, (void *)iv->v, DES_DECRYPT);
-#endif
-
-	vfree (expkey);
-	
-	return res;
+	return evp_crypt(data, key, iv, EVP_des_ede3_cbc(), 0);
 }
 
 int
@@ -1443,40 +1306,14 @@ vchar_t *
 eay_cast_encrypt(data, key, iv)
 	vchar_t *data, *key, *iv;
 {
-	vchar_t *res;
-	CAST_KEY ks;
-
-	CAST_set_key(&ks, key->l, key->v);
-
-	/* allocate buffer for result */
-	if ((res = vmalloc(data->l)) == NULL)
-		return NULL;
-
-	/* decryption data */
-	CAST_cbc_encrypt(data->v, res->v, data->l,
-	                &ks, iv->v, DES_ENCRYPT);
-
-	return res;
+	return evp_crypt(data, key, iv, EVP_cast5_cbc(), 1);
 }
 
 vchar_t *
 eay_cast_decrypt(data, key, iv)
 	vchar_t *data, *key, *iv;
 {
-	vchar_t *res;
-	CAST_KEY ks;
-
-	CAST_set_key(&ks, key->l, key->v);
-
-	/* allocate buffer for result */
-	if ((res = vmalloc(data->l)) == NULL)
-		return NULL;
-
-	/* decryption data */
-	CAST_cbc_encrypt(data->v, res->v, data->l,
-	                &ks, iv->v, DES_DECRYPT);
-
-	return res;
+	return evp_crypt(data, key, iv, EVP_cast5_cbc(), 0);
 }
 
 int
@@ -1519,10 +1356,14 @@ eay_aes_encrypt(data, key, iv)
 
 	/* encryption data */
 	memset(&c, 0, sizeof(c));
-	if (rijndael_cipherInit(&c, MODE_CBC, iv->v) < 0)
+	if (rijndael_cipherInit(&c, MODE_CBC, iv->v) < 0){
+		vfree(res);
 		return NULL;
-	if (rijndael_blockEncrypt(&c, &k, data->v, data->l << 3, res->v) < 0)
+	}
+	if (rijndael_blockEncrypt(&c, &k, data->v, data->l << 3, res->v) < 0){
+		vfree(res);
 		return NULL;
+	}
 
 	return res;
 }
@@ -1545,50 +1386,48 @@ eay_aes_decrypt(data, key, iv)
 
 	/* decryption data */
 	memset(&c, 0, sizeof(c));
-	if (rijndael_cipherInit(&c, MODE_CBC, iv->v) < 0)
+	if (rijndael_cipherInit(&c, MODE_CBC, iv->v) < 0){
+		vfree(res);
 		return NULL;
-	if (rijndael_blockDecrypt(&c, &k, data->v, data->l << 3, res->v) < 0)
+	}
+	if (rijndael_blockDecrypt(&c, &k, data->v, data->l << 3, res->v) < 0){
+		vfree(res);
 		return NULL;
+	}
 
 	return res;
 }
 #else
+static inline const EVP_CIPHER *
+aes_evp_by_keylen(int keylen)
+{
+	switch(keylen) {
+		case 16:
+		case 128:
+			return EVP_aes_128_cbc();
+		case 24:
+		case 192:
+			return EVP_aes_192_cbc();
+		case 32:
+		case 256:
+			return EVP_aes_256_cbc();
+		default:
+			return NULL;
+	}
+}
+
 vchar_t *
 eay_aes_encrypt(data, key, iv)
        vchar_t *data, *key, *iv;
 {
- vchar_t *res;
- AES_KEY ks;
-
- AES_set_encrypt_key(key->v, key->l * 8, &ks);
-
-       /* allocate buffer for result */
-       if ((res = vmalloc(data->l)) == NULL)
-               return NULL;
-
-       /* encryption data */
-       AES_cbc_encrypt(data->v, res->v, data->l,
-                       &ks, iv->v, AES_ENCRYPT);
-  return res;
+	return evp_crypt(data, key, iv, aes_evp_by_keylen(key->l), 1);
 }
 
 vchar_t *
 eay_aes_decrypt(data, key, iv)
        vchar_t *data, *key, *iv;
 {
-  vchar_t *res;
-  AES_KEY ks;
-
-  AES_set_decrypt_key(key->v, key->l * 8, &ks);
-
-       /* allocate buffer for result */
-       if ((res = vmalloc(data->l)) == NULL)
-               return NULL;
-
-       /* decryption data */
-       AES_cbc_encrypt(data->v, res->v, data->l,
-                       &ks, iv->v, AES_DECRYPT);
-  return res;
+	return evp_crypt(data, key, iv, aes_evp_by_keylen(key->l), 0);
 }
 #endif
 
@@ -1699,6 +1538,7 @@ eay_hmacsha2_512_final(c)
 
 	HMAC_Final((HMAC_CTX *)c, res->v, &l);
 	res->l = l;
+	HMAC_cleanup((HMAC_CTX *)c);
 	(void)racoon_free(c);
 
 	if (SHA512_DIGEST_LENGTH != res->l) {
@@ -1755,6 +1595,7 @@ eay_hmacsha2_384_final(c)
 
 	HMAC_Final((HMAC_CTX *)c, res->v, &l);
 	res->l = l;
+	HMAC_cleanup((HMAC_CTX *)c);
 	(void)racoon_free(c);
 
 	if (SHA384_DIGEST_LENGTH != res->l) {
@@ -1811,6 +1652,7 @@ eay_hmacsha2_256_final(c)
 
 	HMAC_Final((HMAC_CTX *)c, res->v, &l);
 	res->l = l;
+	HMAC_cleanup((HMAC_CTX *)c);
 	(void)racoon_free(c);
 
 	if (SHA256_DIGEST_LENGTH != res->l) {
@@ -1868,6 +1710,7 @@ eay_hmacsha1_final(c)
 
 	HMAC_Final((HMAC_CTX *)c, res->v, &l);
 	res->l = l;
+	HMAC_cleanup((HMAC_CTX *)c);
 	(void)racoon_free(c);
 
 	if (SHA_DIGEST_LENGTH != res->l) {
@@ -1924,6 +1767,7 @@ eay_hmacmd5_final(c)
 
 	HMAC_Final((HMAC_CTX *)c, res->v, &l);
 	res->l = l;
+	HMAC_cleanup((HMAC_CTX *)c);
 	(void)racoon_free(c);
 
 	if (MD5_DIGEST_LENGTH != res->l) {
@@ -1988,13 +1832,13 @@ eay_sha2_512_one(data)
 
 	return(res);
 }
-#endif
 
 int
 eay_sha2_512_hashlen()
 {
 	return SHA512_DIGEST_LENGTH << 3;
 }
+#endif
 
 #ifdef WITH_SHA2
 /*
@@ -2048,13 +1892,13 @@ eay_sha2_384_one(data)
 
 	return(res);
 }
-#endif
 
 int
 eay_sha2_384_hashlen()
 {
 	return SHA384_DIGEST_LENGTH << 3;
 }
+#endif
 
 #ifdef WITH_SHA2
 /*
@@ -2108,13 +1952,13 @@ eay_sha2_256_one(data)
 
 	return(res);
 }
-#endif
 
 int
 eay_sha2_256_hashlen()
 {
 	return SHA256_DIGEST_LENGTH << 3;
 }
+#endif
 
 /*
  * SHA functions
@@ -2355,18 +2199,6 @@ end:
 	return(error);
 }
 
-#if 1
-int
-eay_v2bn(bn, var)
-	BIGNUM **bn;
-	vchar_t *var;
-{
-	if ((*bn = BN_bin2bn(var->v, var->l, NULL)) == NULL)
-		return -1;
-
-	return 0;
-}
-#else
 /*
  * convert vchar_t <-> BIGNUM.
  *
@@ -2385,45 +2217,11 @@ eay_v2bn(bn, var)
 	BIGNUM **bn;
 	vchar_t *var;
 {
-	u_char *p;
-	u_char *q;
-	BN_ULONG *r;
-	int l;
-	BN_ULONG num;
+	if ((*bn = BN_bin2bn(var->v, var->l, NULL)) == NULL)
+		return -1;
 
-	*bn = BN_new();
-	if (*bn == NULL)
-		goto err;
-	l = (var->l * 8 + BN_BITS2 - 1) / BN_BITS2;
-	if (bn_expand(*bn, l * BN_BITS2) == NULL)
-		goto err;
-	(*bn)->top = l;
-
-	/* scan from least significant byte */
-	p = (u_char *)var->v;
-	q = (u_char *)(var->v + var->l);
-	r = (*bn)->d;
-	num = 0;
-	l = 0;
-	do {
-		q--;
-		num = num | ((BN_ULONG)*q << (l++ * 8));
-		if (l == BN_BYTES) {
-			*r++ = num;
-			num = 0;
-			l = 0;
-		}
-	} while (p < q);
-	if (l)
-		*r = num;
 	return 0;
-
-err:
-	if (*bn)
-		BN_free(*bn);
-	return -1;
 }
-#endif
 
 int
 eay_bn2v(var, bn)
@@ -2439,6 +2237,182 @@ eay_bn2v(var, bn)
 	return 0;
 }
 
+void
+eay_init()
+{
+	OpenSSL_add_all_algorithms();
+	ERR_load_crypto_strings();
+#ifdef HAVE_OPENSSL_ENGINE_H
+	ENGINE_load_builtin_engines();
+	ENGINE_register_all_complete();
+#endif
+}
+
+vchar_t *
+base64_decode(char *in, long inlen)
+{
+	BIO *bio=NULL, *b64=NULL;
+	vchar_t *res = NULL;
+	char out[inlen*2];
+	long outlen;
+
+	bio = BIO_new_mem_buf(in, inlen);
+	b64 = BIO_new(BIO_f_base64());
+	BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
+	bio = BIO_push(b64, bio);
+
+	outlen = BIO_read(bio, out, inlen * 2);
+	if (outlen <= 0) {
+		plog(LLV_ERROR, LOCATION, NULL, "%s\n", eay_strerror());
+		goto out;
+	}
+
+	res = vmalloc(outlen);
+	if (!res)
+		goto out;
+
+	memcpy(res->v, out, outlen);
+
+out:
+	if (bio)
+		BIO_free_all(bio);
+
+	return res;
+}
+
+vchar_t *
+base64_encode(char *in, long inlen)
+{
+	BIO *bio=NULL, *b64=NULL;
+	char *ptr;
+	long plen = -1;
+	vchar_t *res = NULL;
+
+	bio = BIO_new(BIO_s_mem());
+	b64 = BIO_new(BIO_f_base64());
+	BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
+	bio = BIO_push(b64, bio);
+
+	BIO_write(bio, in, inlen);
+	BIO_flush(bio);
+
+	plen = BIO_get_mem_data(bio, &ptr);
+	res = vmalloc(plen+1);
+	if (!res)
+		goto out;
+	
+	memcpy (res->v, ptr, plen);
+	res->v[plen] = '\0';
+
+out:	
+	if (bio)
+		BIO_free_all(bio);
+
+	return res;
+}
+
+static RSA *
+binbuf_pubkey2rsa(vchar_t *binbuf)
+{
+	BIGNUM *exp, *mod;
+	RSA *rsa_pub = NULL;
+
+	if (binbuf->v[0] > binbuf->l - 1) {
+		plog(LLV_ERROR, LOCATION, NULL, "Plain RSA pubkey format error: decoded string doesn't make sense.\n");
+		goto out;
+	}
+
+	exp = BN_bin2bn(binbuf->v + 1, binbuf->v[0], NULL);
+	mod = BN_bin2bn(binbuf->v + binbuf->v[0] + 1, binbuf->l - binbuf->v[0] - 1, NULL);
+	rsa_pub = RSA_new();
+
+	if (!exp || !mod || !rsa_pub) {
+		plog(LLV_ERROR, LOCATION, NULL, "Plain RSA pubkey parsing error: %s\n", eay_strerror());
+		if (exp)
+			BN_free(exp);
+		if (mod)
+			BN_free(exp);
+		if (rsa_pub)
+			RSA_free(rsa_pub);
+		rsa_pub = NULL;
+		goto out;
+	}
+	
+	rsa_pub->n = mod;
+	rsa_pub->e = exp;
+
+out:
+	return rsa_pub;
+}
+
+RSA *
+base64_pubkey2rsa(char *in)
+{
+	BIGNUM *exp, *mod;
+	RSA *rsa_pub = NULL;
+	vchar_t *binbuf;
+
+	if (strncmp(in, "0s", 2) != 0) {
+		plog(LLV_ERROR, LOCATION, NULL, "Plain RSA pubkey format error: doesn't start with '0s'\n");
+		return NULL;
+	}
+
+	binbuf = base64_decode(in + 2, strlen(in + 2));
+	if (!binbuf) {
+		plog(LLV_ERROR, LOCATION, NULL, "Plain RSA pubkey format error: Base64 decoding failed.\n");
+		return NULL;
+	}
+	
+	if (binbuf->v[0] > binbuf->l - 1) {
+		plog(LLV_ERROR, LOCATION, NULL, "Plain RSA pubkey format error: decoded string doesn't make sense.\n");
+		goto out;
+	}
+
+	rsa_pub = binbuf_pubkey2rsa(binbuf);
+
+out:
+	if (binbuf)
+		vfree(binbuf);
+
+	return rsa_pub;
+}
+
+RSA *
+bignum_pubkey2rsa(BIGNUM *in)
+{
+	RSA *rsa_pub = NULL;
+	vchar_t *binbuf;
+
+	binbuf = vmalloc(BN_num_bytes(in));
+	if (!binbuf) {
+		plog(LLV_ERROR, LOCATION, NULL, "Plain RSA pubkey conversion: memory allocation failed..\n");
+		return NULL;
+	}
+	
+	BN_bn2bin(in, binbuf->v);
+
+	rsa_pub = binbuf_pubkey2rsa(binbuf);
+
+out:
+	if (binbuf)
+		vfree(binbuf);
+
+	return rsa_pub;
+}
+
+u_int32_t
+eay_random()
+{
+	u_int32_t result;
+	vchar_t *vrand;
+
+	vrand = eay_set_random(sizeof(result));
+	memcpy(&result, vrand->v, sizeof(result));
+	vfree(vrand);
+
+	return result;
+}
+
 const char *
 eay_version()
 {
