@@ -1,4 +1,4 @@
-/*	$KAME: handler.h,v 1.44 2002/07/10 23:22:03 itojun Exp $	*/
+/* $Id: handler.h,v 1.11 2004/11/16 15:44:46 ludvigm Exp $ */
 
 /*
  * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
@@ -29,6 +29,17 @@
  * SUCH DAMAGE.
  */
 
+#ifndef _HANDLER_H
+#define _HANDLER_H
+
+#include <sys/queue.h>
+#include <openssl/rsa.h>
+
+#include <sys/time.h>
+
+#include "isakmp_var.h"
+#include "oakley.h"
+
 /* Phase 1 handler */
 /*
  * main mode:
@@ -93,6 +104,9 @@
  * acquire msg		I	R
  * ID payload		I	R		I	R
  */
+#ifdef ENABLE_HYBRID
+struct isakmp_cfg_state;
+#endif
 struct ph1handle {
 	isakmp_index index;
 
@@ -117,6 +131,10 @@ struct ph1handle {
 
 	struct ph1natt_options *natt_options;	/* Selected NAT-T IKE version */
 	u_int32_t natt_flags;		/* NAT-T related flags */
+#ifdef ENABLE_FRAG
+	int frag;			/* IKE phase 1 fragmentation */
+	struct isakmp_frag_item *frag_chain;	/* Received fragments */
+#endif
 
 	struct sched *sce;		/* schedule for expire */
 
@@ -142,6 +160,9 @@ struct ph1handle {
 	cert_t *cert_p;			/* peer's CERT minus general header */
 	cert_t *crl_p;			/* peer's CRL minus general header */
 	cert_t *cr_p;			/* peer's CR not including general */
+	RSA *rsa;			/* my RSA key */
+	RSA *rsa_p;			/* peer's RSA key */
+	struct genlist *rsa_candidates;	/* possible candidates for peer's RSA key */
 	vchar_t *id;			/* ID minus gen header */
 	vchar_t *id_p;			/* partner's ID minus general header */
 					/* i.e. strut ipsecdoi_id_b*. */
@@ -170,11 +191,21 @@ struct ph1handle {
 	struct timeval end;
 #endif
 
+	int		dpd_support;	/* Does remote supports DPD ? */
+	time_t		dpd_lastack;	/* Last ack received */
+	u_int16_t	dpd_seq;		/* DPD seq number to receive */
+	u_int8_t	dpd_fails;		/* number of failures */
+	struct sched	*dpd_r_u;
+
 	u_int32_t msgid2;		/* msgid counter for Phase 2 */
 	int ph2cnt;	/* the number which is negotiated by this phase 1 */
 	LIST_HEAD(_ph2ofph1_, ph2handle) ph2tree;
 
 	LIST_ENTRY(ph1handle) chain;
+#ifdef ENABLE_HYBRID
+	struct isakmp_cfg_state *mode_cfg;	/* ISAKMP mode config state */
+#endif       
+
 };
 
 /* Phase 2 handler */
@@ -388,6 +419,7 @@ extern struct ph1handle *getph1byindex __P((isakmp_index *));
 extern struct ph1handle *getph1byindex0 __P((isakmp_index *));
 extern struct ph1handle *getph1byaddr __P((struct sockaddr *,
 	struct sockaddr *));
+extern struct ph1handle *getph1bydstaddr __P((struct sockaddr *));
 extern vchar_t *dumpph1 __P((void));
 extern struct ph1handle *newph1 __P((void));
 extern void delph1 __P((struct ph1handle *));
@@ -423,3 +455,9 @@ extern int check_recvdpkt __P((struct sockaddr *,
 extern int add_recvdpkt __P((struct sockaddr *, struct sockaddr *,
 	vchar_t *, vchar_t *));
 extern void init_recvdpkt __P((void));
+
+#ifdef ENABLE_HYBRID
+extern int exclude_cfg_addr __P((const struct sockaddr *));
+#endif
+
+#endif /* _HANDLER_H */
