@@ -1,4 +1,4 @@
-/*	$KAME: isakmp.c,v 1.179 2003/11/13 02:30:20 sakane Exp $	*/
+/* $Id: isakmp.c,v 1.32.2.1 2005/01/24 08:50:03 vanhu Exp $ */
 
 /*
  * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
@@ -37,6 +37,7 @@
 #include <sys/queue.h>
 
 #include <netinet/in.h>
+#include <arpa/inet.h>
 
 #include <stdlib.h>
 #include <stdio.h>
@@ -58,10 +59,6 @@
 #endif
 #include <ctype.h>
 
-#if !defined(HAVE_GETADDRINFO) || !defined(HAVE_GETNAMEINFO)
-#include "addrinfo.h"
-#endif
-
 #include "var.h"
 #include "misc.h"
 #include "vmbuf.h"
@@ -76,6 +73,7 @@
 #include "isakmp_var.h"
 #include "isakmp.h"
 #include "oakley.h"
+#include "evt.h"
 #include "handler.h"
 #include "ipsec_doi.h"
 #include "pfkey.h"
@@ -87,17 +85,28 @@
 #include "isakmp_quick.h"
 #include "isakmp_inf.h"
 #include "isakmp_newg.h"
-#include "strnames.h"
-#ifndef HAVE_ARC4RANDOM
-#include "arc4random.h"
+#ifdef ENABLE_HYBRID
+#include "isakmp_xauth.h"
+#include "isakmp_cfg.h"
 #endif
+#ifdef ENABLE_FRAG
+#include "isakmp_frag.h"
+#endif
+#include "strnames.h"
 
 #ifdef ENABLE_NATT
 # include "nattraversal.h"
-# include <linux/udp.h>
-# ifndef SOL_UDP
-#  define SOL_UDP 17
-# endif
+# ifdef __linux__
+#  include <linux/udp.h>
+#  ifndef SOL_UDP
+#   define SOL_UDP 17
+#  endif
+# endif /* __linux__ */
+# if defined(__NetBSD__) || defined(__FreeBSD__)
+#  include <netinet/in.h>
+#  include <netinet/udp.h>
+#  define SOL_UDP IPPROTO_UDP
+# endif /* __NetBSD__ / __FreeBSD__ */
 #endif
 
 static int nostate1 __P((struct ph1handle *, vchar_t *));
@@ -156,6 +165,12 @@ static int isakmp_ph2begin_i __P((struct ph1handle *, struct ph2handle *));
 static int isakmp_ph2begin_r __P((struct ph1handle *, vchar_t *));
 static int etypesw1 __P((int));
 static int etypesw2 __P((int));
+#ifdef ENABLE_FRAG
+static int frag_handler(struct ph1handle *, 
+    vchar_t *, struct sockaddr *, struct sockaddr *);
+#endif
+
+extern char **environ;
 
 /*
  * isakmp packet handler
@@ -208,7 +223,7 @@ isakmp_handler(so_isakmp)
 	/* check isakmp header length, as well as sanity of header length */
 	if (len < sizeof(isakmp) || ntohl(isakmp.len) < sizeof(isakmp)) {
 		plog(LLV_ERROR, LOCATION, (struct sockaddr *)&remote,
-			"packet shorter than isakmp header size (%u, %u, %u)\n",
+			"packet shorter than isakmp header size (%u, %u, %zu)\n",
 			len, ntohl(isakmp.len), sizeof(isakmp));
 		/* dummy receive */
 		if ((len = recvfrom(so_isakmp, (char *)&isakmp, sizeof(isakmp),
@@ -274,7 +289,7 @@ isakmp_handler(so_isakmp)
 	
 	if (len != buf->l) {
 		plog(LLV_ERROR, LOCATION, (struct sockaddr *)&remote,
-			"received invalid length (%d != %d), why ?\n",
+			"received invalid length (%d != %zu), why ?\n",
 			len, buf->l);
 		goto end;
 	}
@@ -528,6 +543,11 @@ isakmp_main(msg, remote, local)
 			return -1;
 		}
 
+#ifdef ENABLE_FRAG
+		if (isakmp->np == ISAKMP_NPTYPE_FRAG)
+			return frag_handler(iph1, msg, remote, local);
+#endif
+
 		/* call main process of phase 1 */
 		if (ph1_main(iph1, msg) < 0) {
 			plog(LLV_ERROR, LOCATION, iph1->remote,
@@ -569,6 +589,11 @@ isakmp_main(msg, remote, local)
 			}
 		}
 
+#ifdef ENABLE_FRAG
+		if (isakmp->np == ISAKMP_NPTYPE_FRAG)
+			return frag_handler(iph1, msg, remote, local);
+#endif
+
 		if (isakmp_info_recv(iph1, msg) < 0)
 			return -1;
 		break;
@@ -588,6 +613,11 @@ isakmp_main(msg, remote, local)
 			return -1;
 		}
 
+#ifdef ENABLE_FRAG
+		if (isakmp->np == ISAKMP_NPTYPE_FRAG)
+			return frag_handler(iph1, msg, remote, local);
+#endif
+
 		/* check status of phase 1 whether negotiated or not. */
 		if (iph1->status != PHASE1ST_ESTABLISHED) {
 			plog(LLV_ERROR, LOCATION, remote,
@@ -636,9 +666,34 @@ isakmp_main(msg, remote, local)
 				"there is no ISAKMP-SA.\n");
 			return -1;
 		}
+
+#ifdef ENABLE_FRAG
+		if (isakmp->np == ISAKMP_NPTYPE_FRAG)
+			return frag_handler(iph1, msg, remote, local);
+#endif
+
 		isakmp_newgroup_r(iph1, msg);
 		break;
 
+#ifdef ENABLE_HYBRID
+	case ISAKMP_ETYPE_CFG:
+		if (iph1 == NULL) {
+			plog(LLV_ERROR, LOCATION, NULL,
+			     "mode config %d from %s, "
+			     "but we have no ISAKMP-SA.\n",
+			     isakmp->etype, saddr2str(remote));
+			return -1;
+		}
+
+#ifdef ENABLE_FRAG
+		if (isakmp->np == ISAKMP_NPTYPE_FRAG)
+			return frag_handler(iph1, msg, remote, local);
+#endif
+
+		isakmp_cfg_r(iph1, msg);
+		break;
+#endif	 
+
 	case ISAKMP_ETYPE_NONE:
 	default:
 		plog(LLV_ERROR, LOCATION, NULL,
@@ -741,6 +796,25 @@ ph1_main(iph1, msg)
 		/* add to the schedule to expire, and seve back pointer. */
 		iph1->sce = sched_new(iph1->approval->lifetime,
 		    isakmp_ph1expire_stub, iph1);
+#ifdef ENABLE_HYBRID
+		if (iph1->mode_cfg->flags & ISAKMP_CFG_VENDORID_XAUTH) {
+			switch(iph1->approval->authmethod) {
+			case OAKLEY_ATTR_AUTH_METHOD_HYBRID_RSA_I:
+			case OAKLEY_ATTR_AUTH_METHOD_HYBRID_DSS_I:
+				xauth_sendreq(iph1);
+				/* XXX Don't process INITIAL_CONTACT */
+				iph1->rmconf->ini_contact = 0;
+				break;
+			default:
+				break;
+			}
+		}
+#endif
+#ifdef ENABLE_DPD
+		/* Schedule the r_u_there.... */
+		if(iph1->dpd_support && iph1->rmconf->dpd_interval)
+			isakmp_sched_r_u(iph1, 0);
+#endif
 
 		/* INITIAL-CONTACT processing */
 		/* don't anything if local test mode. */
@@ -759,6 +833,15 @@ ph1_main(iph1, msg)
 
 		log_ph1established(iph1);
 		plog(LLV_DEBUG, LOCATION, NULL, "===\n");
+
+		/* 
+		 * SA up shell script hook: do it now,except if
+		 * ISAKMP mode config was requested. In the later
+		 * case it is done when we receive the configuration.
+		 */
+		if ((iph1->status == PHASE1ST_ESTABLISHED) &&
+		    !iph1->rmconf->mode_cfg)
+			script_hook(iph1, SCRIPT_PHASE1_UP);	
 	}
 
 	return 0;
@@ -869,6 +952,14 @@ isakmp_ph1begin_i(rmconf, remote, local)
 	iph1->ph2cnt = 0;
 #ifdef HAVE_GSSAPI
 	iph1->gssapi_state = NULL;
+#endif
+#ifdef ENABLE_HYBRID
+	if ((iph1->mode_cfg = isakmp_cfg_mkstate()) == NULL)
+		return -1;
+#endif
+#ifdef ENABLE_FRAG
+	iph1->frag = 0;
+	iph1->frag_chain = NULL;
 #endif
 	iph1->approval = NULL;
 
@@ -968,6 +1059,14 @@ isakmp_ph1begin_r(msg, remote, local, etype)
 	iph1->msgid = 0;
 #ifdef HAVE_GSSAPI
 	iph1->gssapi_state = NULL;
+#endif
+#ifdef ENABLE_HYBRID
+	if ((iph1->mode_cfg = isakmp_cfg_mkstate()) == NULL)
+		return -1;
+#endif
+#ifdef ENABLE_FRAG
+	iph1->frag = 0;
+	iph1->frag_chain = NULL;
 #endif
 	iph1->approval = NULL;
 
@@ -1024,6 +1123,14 @@ isakmp_ph2begin_i(iph1, iph2)
 	struct ph1handle *iph1;
 	struct ph2handle *iph2;
 {
+#ifdef ENABLE_HYBRID
+	if (xauth_check(iph1) != 0) {
+		plog(LLV_ERROR, LOCATION, NULL,
+		    "Attempt to start phase 2 whereas Xauth failed\n");
+		return -1;
+	}
+#endif
+
 	/* found ISAKMP-SA. */
 	plog(LLV_DEBUG, LOCATION, NULL, "===\n");
 	plog(LLV_DEBUG, LOCATION, NULL, "begin QUICK mode.\n");
@@ -1067,6 +1174,13 @@ isakmp_ph2begin_r(iph1, msg)
 #ifdef ENABLE_STATS
 	struct timeval start, end;
 #endif
+#ifdef ENABLE_HYBRID
+	if (xauth_check(iph1) != 0) {
+		plog(LLV_ERROR, LOCATION, NULL,
+		    "Attempt to start phase 2 whereas Xauth failed\n");
+		return -1;
+	}
+#endif
 
 	iph2 = newph2();
 	if (iph2 == NULL) {
@@ -1421,7 +1535,7 @@ isakmp_open()
 			break;
 #ifdef INET6
 		case AF_INET6:
-#ifdef ADVAPI
+#ifdef INET6_ADVAPI
 #ifdef IPV6_RECVPKTINFO
 			pktinfo = IPV6_RECVPKTINFO;
 #else  /* old adv. API */
@@ -1470,21 +1584,31 @@ isakmp_open()
 			saddr2str(p->addr), p->sock);
 
 #ifdef ENABLE_NATT
-		if (p->udp_encap)
-		{
-			int option = UDP_ENCAP_ESPINUDP;
-			if (setsockopt (p->sock, SOL_UDP, UDP_ENCAP,
-					&option, sizeof (option)) < 0) {
-				plog(LLV_ERROR, LOCATION, NULL,
-				     "setsockopt(UDP_ENCAP_ESPINUDP): %s\n",
-				     strerror(errno));
-				goto err_and_next;
-			}
-			else {
-				plog(LLV_INFO, LOCATION, NULL,
-				     "%s used for NAT-T\n",
-				     saddr2str(p->addr));
-				encap_ifnum++;
+		if (p->addr->sa_family == AF_INET) {
+			int option = -1;
+
+
+			if(p->udp_encap)
+				option = UDP_ENCAP_ESPINUDP;
+#if defined(ENABLE_NATT_00) || defined(ENABLE_NATT_01)
+			else
+				option = UDP_ENCAP_ESPINUDP_NON_IKE;
+#endif
+			if(option != -1){
+				if (setsockopt (p->sock, SOL_UDP, UDP_ENCAP,
+								&option, sizeof (option)) < 0) {
+					plog(LLV_ERROR, LOCATION, NULL,
+						 "setsockopt(%s): %s\n",
+						 option == UDP_ENCAP_ESPINUDP ? "UDP_ENCAP_ESPINUDP" : "UDP_ENCAP_ESPINUDP_NON_IKE",
+						 strerror(errno));
+					goto err_and_next;
+				}
+				else {
+					plog(LLV_INFO, LOCATION, NULL,
+						 "%s used for NAT-T\n",
+						 saddr2str(p->addr));
+					encap_ifnum++;
+				}
 			}
 		}
 #endif
@@ -1549,6 +1673,15 @@ isakmp_send(iph1, sbuf)
 #ifdef ENABLE_NATT
 	size_t extralen = NON_ESP_MARKER_USE(iph1) * NON_ESP_MARKER_LEN;
 
+#ifdef ENABLE_FRAG
+	/* 
+	 * Do not add the non ESP marker for a packet that will
+	 * be fragmented. The non ESP marker should appear in 
+	 * all fragment's packets, but not in the fragmented packet
+	 */
+	if (iph1->frag && sbuf->l > ISAKMP_FRAG_MAXLEN) 
+		extralen = 0;
+#endif
 	if (extralen)
 		plog (LLV_DEBUG, LOCATION, NULL, "Adding NON-ESP marker\n");
 
@@ -1571,14 +1704,25 @@ isakmp_send(iph1, sbuf)
 	if (s == -1)
 		return -1;
 
-	plog (LLV_DEBUG, LOCATION, NULL, "%u bytes %s\n", vbuf->l, 
+	plog (LLV_DEBUG, LOCATION, NULL, "%zu bytes %s\n", vbuf->l, 
 	      saddr2str_fromto("from %s to %s", iph1->local, iph1->remote));
 
-	len = sendfromto(s, vbuf->v, vbuf->l,
-			iph1->local, iph1->remote, lcconf->count_persend);
-	if (len == -1) {
-		plog(LLV_ERROR, LOCATION, NULL, "sendfromto failed\n");
-		return -1;
+#ifdef ENABLE_FRAG
+	if (iph1->frag && vbuf->l > ISAKMP_FRAG_MAXLEN) {
+		if (isakmp_sendfrags(iph1, vbuf) == -1) {
+			plog(LLV_ERROR, LOCATION, NULL, 
+			    "isakmp_sendfrags failed\n");
+			return -1;
+		}
+	} else 
+#endif
+	{
+		len = sendfromto(s, vbuf->v, vbuf->l,
+		    iph1->local, iph1->remote, lcconf->count_persend);
+		if (len == -1) {
+			plog(LLV_ERROR, LOCATION, NULL, "sendfromto failed\n");
+			return -1;
+		}
 	}
 
 	return 0;
@@ -1600,6 +1744,8 @@ isakmp_ph1resend(iph1)
 		plog(LLV_ERROR, LOCATION, NULL,
 			"phase1 negotiation failed due to time up. %s\n",
 			isakmp_pindex(&iph1->index, iph1->msgid));
+		EVT_PUSH(iph1->local, iph1->remote, 
+		    EVTT_PEER_NO_RESPONSE, NULL);
 
 		remph1(iph1);
 		delph1(iph1);
@@ -1638,6 +1784,7 @@ isakmp_ph2resend(iph2)
 		plog(LLV_ERROR, LOCATION, NULL,
 			"phase2 negotiation failed due to time up. %s\n",
 				isakmp_pindex(&iph2->ph1->index, iph2->msgid));
+		EVT_PUSH(iph2->src, iph2->dst, EVTT_PEER_NO_RESPONSE, NULL);
 		unbindph12(iph2);
 		remph2(iph2);
 		delph2(iph2);
@@ -1727,6 +1874,7 @@ isakmp_ph1delete(iph1)
 	plog(LLV_INFO, LOCATION, NULL,
 		"ISAKMP-SA deleted %s-%s spi:%s\n",
 		src, dst, isakmp_pindex(&iph1->index, 0));
+	EVT_PUSH(iph1->local, iph1->remote, EVTT_PHASE1_DOWN, NULL);
 	racoon_free(src);
 	racoon_free(dst);
 
@@ -2201,7 +2349,7 @@ isakmp_newmsgid2(iph1)
 	u_int32_t msgid2;
 
 	do {
-		msgid2 = arc4random();
+		msgid2 = eay_random();
 	} while (getph2bymsgid(iph1, msgid2));
 
 	return msgid2;
@@ -2274,7 +2422,7 @@ set_isakmp_payload(buf, src, nptype)
 	struct isakmp_gen *gen;
 	caddr_t p = buf;
 
-	plog(LLV_DEBUG, LOCATION, NULL, "add payload of len %d, next type %d\n",
+	plog(LLV_DEBUG, LOCATION, NULL, "add payload of len %zu, next type %d\n",
 	    src->l, nptype);
 
 	gen = (struct isakmp_gen *)p;
@@ -2589,6 +2737,7 @@ log_ph1established(iph1)
 		"ISAKMP-SA established %s-%s spi:%s\n",
 		src, dst,
 		isakmp_pindex(&iph1->index, 0));
+	EVT_PUSH(iph1->local, iph1->remote, EVTT_PHASE1_UP, NULL);
 	racoon_free(src);
 	racoon_free(dst);
 
@@ -2662,3 +2811,77 @@ isakmp_plist_set_all (struct payload_list **plist, struct ph1handle *iph1)
 end:
 	return NULL;
 }
+
+#ifdef ENABLE_FRAG
+int 
+frag_handler(iph1, msg, remote, local)
+	struct ph1handle *iph1;
+	vchar_t *msg;
+	struct sockaddr *remote;
+	struct sockaddr *local;
+{
+	vchar_t *newmsg;
+
+	if (isakmp_frag_extract(iph1, msg) == 1) {
+		if ((newmsg = isakmp_frag_reassembly(iph1)) == NULL) {
+			plog(LLV_ERROR, LOCATION, remote, 
+			    "Packet reassembly failed\n");
+			return -1;
+		}
+		return isakmp_main(newmsg, remote, local);
+	}
+
+	return 0;
+}
+#endif
+
+void
+script_hook(iph1, script)
+	struct ph1handle *iph1;
+	int script;
+{
+#define IP_MAX 40
+	char addrstr[IP_MAX];
+	char *argv[3];
+	struct sockaddr_in *sin;
+
+	if (iph1->rmconf->script[script] == NULL)
+		return;
+
+#ifdef ENABLE_HYBRID
+	isakmp_cfg_setenv(iph1);
+#endif
+
+	/* local address */
+	sin = (struct sockaddr_in *)iph1->local;
+	inet_ntop(sin->sin_family, &sin->sin_addr, addrstr, IP_MAX);
+	setenv("LOCAL_ADDR", addrstr, 1);
+
+
+	/* Peer address */
+	sin = (struct sockaddr_in *)iph1->remote;
+	inet_ntop(sin->sin_family, &sin->sin_addr, addrstr, IP_MAX);
+	setenv("REMOTE_ADDR", addrstr, 1);
+
+	argv[0] = iph1->rmconf->script[script]->v;
+	argv[1] = script_names[script];
+	argv[2] = NULL;
+
+	switch (fork()) {
+	case 0:
+		execve(argv[0], argv, environ);
+		plog(LLV_ERROR, LOCATION, NULL, "execve(\"%s\") failed: %s\n",
+		    argv[0], strerror(errno));
+		return;
+		break;
+	case -1:
+		plog(LLV_ERROR, LOCATION, NULL, "Cannot fork: %s\n",
+		    strerror(errno));
+		return;
+		break;
+	default:
+		break;
+	}
+
+	return;
+}
