@@ -1,4 +1,4 @@
-/*	$KAME: isakmp_inf.c,v 1.83 2004/01/14 09:15:00 itojun Exp $	*/
+/* $Id: isakmp_inf.c,v 1.14 2004/12/07 13:12:25 vanhu Exp $ */
 
 /*
  * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
@@ -37,8 +37,8 @@
 
 #include <net/pfkeyv2.h>
 #include <netinet/in.h>
-#ifdef IPV6_INRIA_VERSION
 #include <sys/queue.h>
+#ifndef HAVE_NETINET6_IPSEC
 #include <netinet/ipsec.h>
 #else
 #include <netinet6/ipsec.h>
@@ -73,7 +73,12 @@
 #include "remoteconf.h"
 #include "sockmisc.h"
 #include "isakmp_var.h"
+#include "evt.h"
 #include "isakmp.h"
+#ifdef ENABLE_HYBRID
+#include "isakmp_xauth.h"
+#include "isakmp_cfg.h" 
+#endif
 #include "isakmp_inf.h"
 #include "oakley.h"
 #include "handler.h"
@@ -90,6 +95,15 @@
 static int isakmp_info_recv_n __P((struct ph1handle *, vchar_t *));
 static int isakmp_info_recv_d __P((struct ph1handle *, vchar_t *));
 
+#ifdef ENABLE_DPD
+static int isakmp_info_recv_r_u __P((struct ph1handle *,
+	struct isakmp_pl_ru *, u_int32_t));
+static int isakmp_info_recv_r_u_ack __P((struct ph1handle *,
+	struct isakmp_pl_ru *, u_int32_t));
+static void isakmp_info_send_r_u __P((void *));
+static void purge_remote __P((struct ph1handle *));
+#endif
+
 static void purge_isakmp_spi __P((int, isakmp_index *, size_t));
 static void purge_ipsec_spi __P((struct sockaddr *, int, u_int32_t *, size_t));
 static void info_recv_initialcontact __P((struct ph1handle *));
@@ -412,6 +426,14 @@ isakmp_info_send_nx(isakmp, remote, local, type, data)
 	iph1->version = isakmp->v;
 	iph1->flags = 0;
 	iph1->msgid = 0;	/* XXX */
+#ifdef ENABLE_HYBRID
+	if ((iph1->mode_cfg = isakmp_cfg_mkstate()) == NULL)
+		return -1;
+#endif
+#ifdef ENABLE_FRAG
+	iph1->frag = 0;
+	iph1->frag_chain = NULL;
+#endif
 
 	/* copy remote address */
 	if (copy_ph1addresses(iph1, rmconf, remote, local) < 0)
@@ -823,6 +845,17 @@ isakmp_info_recv_n(iph1, msg)
 	case ISAKMP_NTYPE_INITIAL_CONTACT:
 		info_recv_initialcontact(iph1);
 		break;
+#ifdef ENABLE_DPD
+	case ISAKMP_NTYPE_R_U_THERE:
+		isakmp_info_recv_r_u(iph1, (struct isakmp_pl_ru *)n,
+				     ((struct isakmp *)msg->v)->msgid);
+		break;
+	case ISAKMP_NTYPE_R_U_THERE_ACK:
+		isakmp_info_recv_r_u_ack(iph1, (struct isakmp_pl_ru *)n,
+					 ((struct isakmp *)msg->v)->msgid);
+		break;
+#endif
+
 	default:
 	    {
 		u_int32_t msgid = ((struct isakmp *)msg->v)->msgid;
@@ -1244,7 +1277,17 @@ isakmp_info_recv_d(iph1, msg)
 			plog(LLV_ERROR, LOCATION, iph1->remote,
 				"delete payload with invalid doi:%d.\n",
 				ntohl(d->doi));
+#ifdef ENABLE_HYBRID
+			/*
+			 * At deconnexion time, Cisco VPN client does this
+			 * with a zero DOI. Don't give up in that situation.
+			 */
+			if (((iph1->mode_cfg->flags &
+			    ISAKMP_CFG_VENDORID_UNITY) == 0) || (d->doi != 0))
+				continue;
+#else
 			continue;
+#endif
 		}
 
 		num_spi = ntohs(d->num_spi);
@@ -1279,6 +1322,8 @@ isakmp_info_recv_d(iph1, msg)
 					d->spi_size, d->proto_id);
 				continue;
 			}
+			EVT_PUSH(iph1->local, iph1->remote, 
+			    EVTT_PEER_DELETE, NULL);
 			purge_ipsec_spi(iph1->remote, d->proto_id,
 			    (u_int32_t *)(d + 1), num_spi);
 			break;
@@ -1346,6 +1391,10 @@ isakmp_check_notify(gen, iph1)
 			"ignore INITIAL-CONTACT notification, "
 			"because it is only accepted after phase1.\n");
 		break;
+	case ISAKMP_NTYPE_HEARTBEAT:
+		plog(LLV_WARNING, LOCATION, iph1->remote,
+			"ignore HEARTBEAT notification\n");
+		break;
 	default:
 		isakmp_info_send_n1(iph1, ISAKMP_NTYPE_INVALID_PAYLOAD_TYPE, NULL);
 		plog(LLV_ERROR, LOCATION, iph1->remote,
@@ -1356,3 +1405,299 @@ isakmp_check_notify(gen, iph1)
 	return;
 }
 
+
+#ifdef ENABLE_DPD
+static int
+isakmp_info_recv_r_u (iph1, ru, msgid)
+	struct ph1handle *iph1;
+	struct isakmp_pl_ru *ru;
+	u_int32_t msgid;
+{
+	struct isakmp_pl_ru *ru_ack;
+	vchar_t *payload = NULL;
+	int tlen;
+	int error = 0;
+
+	plog(LLV_DEBUG, LOCATION, iph1->remote,
+		 "DPD R-U-There received\n");
+
+	/* XXX should compare cookies with iph1->index?
+	   Or is this already done by calling function?  */
+	tlen = sizeof(*ru_ack);
+	payload = vmalloc(tlen);
+	if (payload == NULL) { 
+		plog(LLV_ERROR, LOCATION, NULL,
+			"failed to get buffer to send.\n");
+		return errno;
+	}
+
+	ru_ack = (struct isakmp_pl_ru *)payload->v;
+	ru_ack->h.np = ISAKMP_NPTYPE_NONE;
+	ru_ack->h.len = htons(tlen);
+	ru_ack->doi = htonl(IPSEC_DOI);
+	ru_ack->type = htons(ISAKMP_NTYPE_R_U_THERE_ACK);
+	ru_ack->proto_id = IPSECDOI_PROTO_ISAKMP; /* XXX ? */
+	ru_ack->spi_size = sizeof(isakmp_index);
+	memcpy(ru_ack->i_ck, ru->i_ck, sizeof(cookie_t));
+	memcpy(ru_ack->r_ck, ru->r_ck, sizeof(cookie_t));	
+	ru_ack->data = ru->data;
+
+	/* XXX Should we do FLAG_A ?  */
+	error = isakmp_info_send_common(iph1, payload, ISAKMP_NPTYPE_N,
+					ISAKMP_FLAG_E);
+	vfree(payload);
+
+	plog(LLV_DEBUG, LOCATION, NULL, "received a valid R-U-THERE, ACK sent\n");
+
+	/* Should we mark tunnel as active ? */
+	return error;
+}
+
+static int
+isakmp_info_recv_r_u_ack (iph1, ru, msgid)
+	struct ph1handle *iph1;
+	struct isakmp_pl_ru *ru;
+	u_int32_t msgid;
+{
+
+	plog(LLV_DEBUG, LOCATION, iph1->remote,
+		 "DPD R-U-There-Ack received\n");
+
+	/* XXX Maintain window of acceptable sequence numbers ?
+	 * => ru->data <= iph2->dpd_seq &&
+	 *    ru->data >= iph2->dpd_seq - iph2->dpd_fails ? */
+	if (ntohl(ru->data) != iph1->dpd_seq-1) {
+		plog(LLV_ERROR, LOCATION, iph1->remote,
+			 "Wrong DPD sequence number (%d, %d expected).\n", 
+			 ntohl(ru->data), iph1->dpd_seq-1);
+		return 0;
+	}
+
+	if (memcmp(ru->i_ck, iph1->index.i_ck, sizeof(cookie_t)) ||
+	    memcmp(ru->r_ck, iph1->index.r_ck, sizeof(cookie_t))) {
+		plog(LLV_ERROR, LOCATION, iph1->remote,
+			 "Cookie mismatch in DPD ACK!.\n");
+		return 0;
+	}
+
+	iph1->dpd_fails = 0;
+
+	/* Useless ??? */
+	iph1->dpd_lastack = time(NULL);
+
+	if (iph1->dpd_r_u != NULL)
+		SCHED_KILL(iph1->dpd_r_u);
+
+	isakmp_sched_r_u(iph1, 0);
+
+	plog(LLV_DEBUG, LOCATION, NULL, "received an R-U-THERE-ACK\n");
+
+	return 0;
+}
+
+
+
+static void
+purge_remote(iph1)
+	struct ph1handle *iph1;
+{
+	vchar_t *buf = NULL;
+	struct sadb_msg *msg, *next, *end;
+	struct sadb_sa *sa;
+	struct sockaddr *src, *dst;
+	caddr_t mhp[SADB_EXT_MAX + 1];
+
+	/* Delete all phase2 SAs */
+	buf = pfkey_dump_sadb(SADB_SATYPE_UNSPEC);
+	if (buf == NULL) {
+		plog(LLV_DEBUG, LOCATION, NULL,
+			"pfkey_dump_sadb returned nothing.\n");
+		return;
+	}
+
+	msg = (struct sadb_msg *)buf->v;
+	end = (struct sadb_msg *)(buf->v + buf->l);
+
+	while (msg < end) {
+		if ((msg->sadb_msg_len << 3) < sizeof(*msg))
+			break;
+		next = (struct sadb_msg *)((caddr_t)msg + (msg->sadb_msg_len << 3));
+		if (msg->sadb_msg_type != SADB_DUMP) {
+			msg = next;
+			continue;
+		}
+
+		if (pfkey_align(msg, mhp) || pfkey_check(mhp)) {
+			plog(LLV_ERROR, LOCATION, NULL,
+				"pfkey_check (%s)\n", ipsec_strerror());
+			msg = next;
+			continue;
+		}
+
+		sa = (struct sadb_sa *)(mhp[SADB_EXT_SA]);
+		if (!sa ||
+		    !mhp[SADB_EXT_ADDRESS_SRC] ||
+		    !mhp[SADB_EXT_ADDRESS_DST]) {
+			msg = next;
+			continue;
+		}
+		src = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_SRC]);
+		dst = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_DST]);
+
+		if (sa->sadb_sa_state != SADB_SASTATE_MATURE &&
+		    sa->sadb_sa_state != SADB_SASTATE_DYING) {
+			msg = next;
+			continue;
+		}
+
+		/* delete in/outbound SAs */
+		if (cmpsaddrwop(iph1->remote, dst) &&
+		    cmpsaddrwop(iph1->remote, src)) {
+			msg = next;
+			continue;
+		}
+
+		pfkey_send_delete(lcconf->sock_pfkey,
+				  msg->sadb_msg_satype,
+				  IPSEC_MODE_ANY,
+				  src, dst, sa->sadb_sa_spi);
+
+		plog(LLV_INFO, LOCATION, NULL,
+			 "purged IPsec-SA spi=%u.\n",
+			 ntohl(sa->sadb_sa_spi));
+
+		msg = next;
+	}
+
+	if (buf)
+		vfree(buf);
+
+	/* Mark the phase1 handler as EXPIRED */
+	plog(LLV_INFO, LOCATION, NULL,
+		 "purged ISAKMP-SA spi=%s.\n",
+		 isakmp_pindex(&(iph1->index), iph1->msgid));
+
+	if (iph1->sce)
+		SCHED_KILL(iph1->sce);
+
+	iph1->status = PHASE1ST_EXPIRED;
+	iph1->sce = sched_new(1, isakmp_ph1delete_stub, iph1);
+}
+
+/*
+ * send Delete payload (for ISAKMP SA) in Informational exchange.
+ */
+static void
+isakmp_info_send_r_u(arg)
+	void *arg;
+{
+	struct ph1handle *iph1 = arg;
+
+	/* create R-U-THERE payload */
+	struct isakmp_pl_ru *ru;
+	vchar_t *payload = NULL;
+	int tlen;
+	int error = 0;
+
+	plog(LLV_DEBUG, LOCATION, iph1->remote, "DPD monitoring....\n");
+
+	if (iph1->dpd_fails >= iph1->rmconf->dpd_maxfails) {
+		EVT_PUSH(iph1->local, iph1->remote, EVTT_DPD_TIMEOUT, NULL);
+		purge_remote(iph1);
+		plog(LLV_DEBUG, LOCATION, iph1->remote,
+			 "DPD: remote seems to be dead\n");
+
+		/* Do not reschedule here: phase1 is deleted,
+		 * DPD will be reactivated when a new ph1 will be negociated
+		 */
+		return;
+	}
+
+	/* TODO: check recent activity to avoid useless sends... */
+
+	/* XXX: why do we have a NULL LIST_FIRST even when a Phase2 exists ??? */
+#if 0
+	if (LIST_FIRST(&iph1->ph2tree) == NULL){
+		/* XXX: No Ph2 => no need to test ph1 ?
+		 */
+		/* Reschedule the r_u_there....
+		   XXX: reschedule when a new ph2 ?
+		 */
+		isakmp_sched_r_u(iph1, 0);
+		plog(LLV_DEBUG, LOCATION, iph1->remote,
+			 "no phase2 handler, rescheduling send_r_u (%d).\n", iph1->rmconf->dpd_interval);
+		return 0;
+	}
+#endif
+
+	tlen = sizeof(*ru);
+	payload = vmalloc(tlen);
+	if (payload == NULL) {
+		plog(LLV_ERROR, LOCATION, NULL, 
+			 "failed to get buffer for payload.\n");
+		return;
+	}
+	ru = (struct isakmp_pl_ru *)payload->v;
+	ru->h.np = ISAKMP_NPTYPE_NONE;
+	ru->h.len = htons(tlen);
+	ru->doi = htonl(IPSEC_DOI);
+	ru->type = htons(ISAKMP_NTYPE_R_U_THERE);
+	ru->proto_id = IPSECDOI_PROTO_ISAKMP; /* XXX ?*/
+	ru->spi_size = sizeof(isakmp_index);
+
+	memcpy(ru->i_ck, iph1->index.i_ck, sizeof(cookie_t));
+	memcpy(ru->r_ck, iph1->index.r_ck, sizeof(cookie_t));
+
+	if (iph1->dpd_seq == 0){
+		/* generate a random seq which is not too big */
+		srand(time(NULL));
+		iph1->dpd_seq = rand() & 0x0fff;
+	}
+
+	ru->data = htonl(iph1->dpd_seq);
+
+	error = isakmp_info_send_common(iph1, payload, ISAKMP_NPTYPE_N, 0);
+	vfree(payload);
+
+	plog(LLV_DEBUG, LOCATION, iph1->remote,
+		 "DPD R-U-There sent (%d)\n", error);
+
+	/* will be decreased if ACK received... */
+	iph1->dpd_fails++;
+
+	/* XXX should be increased only when ACKed ? */
+	iph1->dpd_seq++;
+
+	/* Reschedule the r_u_there with a short delay,
+	 * will be deleted/rescheduled if ACK received before */
+	isakmp_sched_r_u(iph1, 1);
+
+	plog(LLV_DEBUG, LOCATION, iph1->remote,
+		 "rescheduling send_r_u (%d).\n", iph1->rmconf->dpd_retry);
+}
+
+/* Schedule a new R-U-THERE */
+int
+isakmp_sched_r_u(iph1, retry)
+	struct ph1handle *iph1;
+	int retry;
+{
+	if(iph1 == NULL ||
+	   iph1->rmconf == NULL)
+		return 1;
+
+
+	if(iph1->dpd_support == 0 ||
+	   iph1->rmconf->dpd_interval == 0)
+		return 0;
+
+	if(retry)
+		iph1->dpd_r_u = sched_new(iph1->rmconf->dpd_retry,
+					  isakmp_info_send_r_u, iph1);
+	else
+		sched_new(iph1->rmconf->dpd_interval,
+			  isakmp_info_send_r_u, iph1);
+
+	return 0;
+}
+#endif
