@@ -1,4 +1,4 @@
-/*	$KAME: isakmp_quick.c,v 1.95 2003/10/21 07:18:03 itojun Exp $	*/
+/* $Id: isakmp_quick.c,v 1.12.2.1 2005/01/11 08:51:40 manubsd Exp $ */
 
 /*
  * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
@@ -52,7 +52,7 @@
 # endif
 #endif
 
-#ifdef IPV6_INRIA_VERSION
+#ifndef HAVE_NETINET6_IPSEC
 #include <netinet/ipsec.h>
 #else
 #include <netinet6/ipsec.h>
@@ -474,7 +474,7 @@ quick_i2recv(iph2, msg0)
 	/* Fixed buffer for calculating HASH */
 	memcpy(hbuf->v, iph2->nonce->v, iph2->nonce->l);
 	plog(LLV_DEBUG, LOCATION, NULL,
-		"HASH allocated:hbuf->l=%d actual:tlen=%d\n",
+		"HASH allocated:hbuf->l=%zu actual:tlen=%zu\n",
 		hbuf->l, tlen + iph2->nonce->l);
 	/* adjust buffer length for HASH */
 	hbuf->l = iph2->nonce->l + tlen;
@@ -1512,7 +1512,7 @@ quick_r3send(iph2, msg0)
 	n = (struct isakmp_pl_n *)notify->v;
 	n->h.np = ISAKMP_NPTYPE_NONE;
 	n->h.len = htons(tlen);
-	n->doi = IPSEC_DOI;
+	n->doi = htons(IPSEC_DOI);
 	n->proto_id = iph2->approval->head->proto_id;
 	n->spi_size = sizeof(iph2->approval->head->spisize);
 	n->type = htons(ISAKMP_NTYPE_CONNECTED);
@@ -1579,6 +1579,18 @@ end:
 	return error;
 }
 
+static int
+tunnel_mode_prop(p)
+	struct saprop *p;
+{
+	struct saproto *pr;
+
+	for (pr = p->head; pr; pr = pr->next)
+		if (pr->encmode == IPSECDOI_ATTR_ENC_MODE_TUNNEL)
+			return 1;
+	return 0;
+}
+
 /*
  * set SA to kernel.
  */
@@ -1648,10 +1660,24 @@ quick_r3prep(iph2, msg0)
 		plog(LLV_DEBUG, LOCATION, NULL,
 			"pfkey spdupdate2(inbound) sent.\n");
 
+		spidx = (struct policyindex *)iph2->spidx_gen;
+#ifdef HAVE_POLICY_FWD
+		/* make forward policy if required */
+		if (tunnel_mode_prop(iph2->approval)) {
+			spidx->dir = IPSEC_DIR_FWD;
+			if (pk_sendspdupdate2(iph2) < 0) {
+				plog(LLV_ERROR, LOCATION, NULL,
+					"pfkey spdupdate2(forward) failed.\n");
+				goto end;
+			}
+			plog(LLV_DEBUG, LOCATION, NULL,
+				"pfkey spdupdate2(forward) sent.\n");
+		}
+#endif
+
 		/* make outbound policy */
 		iph2->src = src;
 		iph2->dst = dst;
-		spidx = (struct policyindex *)iph2->spidx_gen;
 		spidx->dir = IPSEC_DIR_OUTBOUND;
 		addr = spidx->src;
 		spidx->src = spidx->dst;
@@ -2032,6 +2058,20 @@ get_proposal_r(iph2)
 			"no policy found: %s\n", spidx2str(&spidx));
 		return ISAKMP_INTERNAL_ERROR;
 	}
+	/* Refresh existing generated policies
+	 */
+	if (iph2->ph1->rmconf->gen_policy) {
+		plog(LLV_INFO, LOCATION, NULL,
+			 "Update the generated policy : %s\n",
+			 spidx2str(&spidx));
+		iph2->spidx_gen = racoon_malloc(sizeof(spidx));
+		if (!iph2->spidx_gen) {
+			plog(LLV_ERROR, LOCATION, NULL,
+				 "buffer allocation failed.\n");
+			return ISAKMP_INTERNAL_ERROR;
+		}
+		memcpy(iph2->spidx_gen, &spidx, sizeof(spidx));
+	}
 
 	/* get outbound policy */
     {
