@@ -1,4 +1,4 @@
-/*	$KAME: pfkey.c,v 1.142 2003/12/16 01:13:30 suz Exp $	*/
+/* $Id: pfkey.c,v 1.27.2.2 2005/01/29 09:36:26 manubsd Exp $ */
 
 /*
  * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
@@ -29,10 +29,29 @@
  * SUCH DAMAGE.
  */
 
-#define _PFKEY_C_
-
 #include "config.h"
 
+#include <stdlib.h>
+#include <string.h>
+#include <stdio.h>
+#include <netdb.h>
+#include <errno.h>
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+#include <netdb.h>
+#include <netinet/in.h>
+#include <arpa/inet.h>
+
+#ifdef ENABLE_NATT
+# ifdef __linux__
+#  include <linux/udp.h>
+# endif
+# if defined(__NetBSD__) || defined(__FreeBSD__)
+#  include <netinet/udp.h>
+# endif
+#endif
+
 #include <sys/types.h>
 #include <sys/param.h>
 #include <sys/socket.h>
@@ -43,25 +62,12 @@
 #include <net/pfkeyv2.h>
 
 #include <netinet/in.h>
-#ifdef IPV6_INRIA_VERSION
+#ifndef HAVE_NETINET6_IPSEC
 #include <netinet/ipsec.h>
 #else
 #include <netinet6/ipsec.h>
 #endif
 
-#include <linux/udp.h>
-
-#include <stdlib.h>
-#include <string.h>
-#include <stdio.h>
-#include <netdb.h>
-#include <errno.h>
-#ifdef HAVE_UNISTD_H
-#include <unistd.h>
-#endif
-#include <netdb.h>
-#include <arpa/inet.h>
-
 #include "libpfkey.h"
 
 #include "var.h"
@@ -89,10 +95,13 @@
 #include "strnames.h"
 #include "backupsa.h"
 #include "gcmalloc.h"
-#ifndef HAVE_ARC4RANDOM
-#include "arc4random.h"
-#endif
 #include "nattraversal.h"
+#include "crypto_openssl.h"
+#include "grabmyaddr.h"
+
+#if defined(SADB_X_EALG_RIJNDAELCBC) && !defined(SADB_X_EALG_AESCBC)
+#define SADB_X_EALG_AESCBC  SADB_X_EALG_RIJNDAELCBC
+#endif
 
 /* prototype */
 static u_int ipsecdoi2pfkey_aalg __P((u_int));
@@ -331,6 +340,7 @@ done:
 	return buf;
 }
 
+#ifdef ENABLE_ADMINPORT
 /*
  * flush SADB
  */
@@ -353,6 +363,7 @@ pfkey_flush_sadb(proto)
 
 	return;
 }
+#endif
 
 /*
  * These are the SATYPEs that we manage.  We register to get
@@ -378,7 +389,7 @@ pfkey_init()
 
 	if ((lcconf->sock_pfkey = pfkey_open()) < 0) {
 		plog(LLV_ERROR, LOCATION, NULL,
-			"libipsec failed pfkey open (%s)", ipsec_strerror());
+			"libipsec failed pfkey open (%s)\n", ipsec_strerror());
 		return -1;
 	}
 
@@ -390,7 +401,7 @@ pfkey_init()
 					pfkey_satypes[i].ps_satype) < 0 ||
 		    pfkey_recv_register(lcconf->sock_pfkey) < 0) {
 			plog(LLV_WARNING, LOCATION, NULL,
-			    "failed to register %s (%s)",
+			    "failed to register %s (%s)\n",
 			    pfkey_satypes[i].ps_name,
 			    ipsec_strerror());
 			reg_fail++;
@@ -399,7 +410,7 @@ pfkey_init()
 
 	if (reg_fail == pfkey_nsatypes) {
 		plog(LLV_ERROR, LOCATION, NULL,
-			"failed to regist any protocol.");
+			"failed to regist any protocol.\n");
 		pfkey_close(lcconf->sock_pfkey);
 		return -1;
 	}
@@ -408,7 +419,7 @@ pfkey_init()
 
 	if (pfkey_send_spddump(lcconf->sock_pfkey) < 0) {
 		plog(LLV_ERROR, LOCATION, NULL,
-			"libipsec sending spddump failed: %s",
+			"libipsec sending spddump failed: %s\n",
 			ipsec_strerror());
 		pfkey_close(lcconf->sock_pfkey);
 		return -1;
@@ -1035,11 +1046,12 @@ pk_sendupdate(iph2)
 #else
 		lifebyte = 0;
 #endif
-		plog(LLV_DEBUG, LOCATION, NULL, "call pfkey_send_update_nat\n");
 
+#ifdef ENABLE_NATT
+		plog(LLV_DEBUG, LOCATION, NULL, "call pfkey_send_update_nat\n");
 		if (pr->udp_encap) {
 			memset (&natt, 0, sizeof (natt));
-			natt.type = UDP_ENCAP_ESPINUDP;
+			natt.type = iph2->ph1->natt_options->encaps_type;
 			natt.sport = extract_port (iph2->ph1->remote);
 			natt.dport = extract_port (iph2->ph1->local);
 			natt.oa = NULL;		// FIXME: Here comes OA!!!
@@ -1061,11 +1073,32 @@ pk_sendupdate(iph2)
 				0, lifebyte, iph2->approval->lifetime, 0,
 				iph2->seq,
 				natt.type, natt.sport, natt.dport, natt.oa) < 0) {
+			plog(LLV_ERROR, LOCATION, NULL,
+				"libipsec failed send update_nat (%s)\n",
+				ipsec_strerror());
+			return -1;
+		}
+#else
+		plog(LLV_DEBUG, LOCATION, NULL, "call pfkey_send_update\n");
+		if (pfkey_send_update(
+				lcconf->sock_pfkey,
+				satype,
+				mode,
+				dst,
+				src,
+				pr->spi,
+				pr->reqid_in,
+				wsize,	
+				pr->keymat->v,
+				e_type, e_keylen, a_type, a_keylen, flags,
+				0, lifebyte, iph2->approval->lifetime, 0,
+				iph2->seq) < 0) {
 			plog(LLV_ERROR, LOCATION, NULL,
 				"libipsec failed send update (%s)\n",
 				ipsec_strerror());
 			return -1;
 		}
+#endif /* ENABLE_NATT */
 
 		if (!lcconf->pathinfo[LC_PATHTYPE_BACKUPSA])
 			continue;
@@ -1304,6 +1337,7 @@ pk_sendadd(iph2)
 		lifebyte = 0;
 #endif
 
+#ifdef ENABLE_NATT
 		plog(LLV_DEBUG, LOCATION, NULL, "call pfkey_send_add_nat\n");
 
 		if (pr->udp_encap) {
@@ -1330,11 +1364,33 @@ pk_sendadd(iph2)
 				0, lifebyte, iph2->approval->lifetime, 0,
 				iph2->seq,
 				natt.type, natt.sport, natt.dport, natt.oa) < 0) {
+			plog(LLV_ERROR, LOCATION, NULL,
+				"libipsec failed send add_nat (%s)\n",
+				ipsec_strerror());
+			return -1;
+		}
+#else
+		plog(LLV_DEBUG, LOCATION, NULL, "call pfkey_send_add\n");
+
+		if (pfkey_send_add(
+				lcconf->sock_pfkey,
+				satype,
+				mode,
+				src,
+				dst,
+				pr->spi_p,
+				pr->reqid_out,
+				wsize,
+				pr->keymat_p->v,
+				e_type, e_keylen, a_type, a_keylen, flags,
+				0, lifebyte, iph2->approval->lifetime, 0,
+				iph2->seq) < 0) {
 			plog(LLV_ERROR, LOCATION, NULL,
 				"libipsec failed send add (%s)\n",
 				ipsec_strerror());
 			return -1;
 		}
+#endif /* ENABLE_NATT */
 
 		if (!lcconf->pathinfo[LC_PATHTYPE_BACKUPSA])
 			continue;
@@ -1592,6 +1648,34 @@ pk_recvacquire(mhp)
 		return 0;
 	}
     }
+   	
+    	/* ignore, if we do not listen on source address */
+	{
+		/* reasons behind:
+		 * - if we'll contact peer from address we do not listen -
+		 *   we will be unable to complete negotiation;
+		 * - if we'll negotiate using address we're listening -
+		 *   remote peer will send packets to address different
+		 *   than one in the policy, so kernel will drop them;
+		 * => therefore this acquire is not for us! --Aidas
+		 */
+		struct sockaddr *sa = PFKEY_ADDR_SADDR(mhp[SADB_EXT_ADDRESS_SRC]);
+		struct myaddrs *p;
+		int do_listen = 0;
+		for (p = lcconf->myaddrs; p; p = p->next) {
+			if (!cmpsaddrwop(p->addr, sa)) {
+				do_listen = 1;
+				break;
+			}
+		}
+
+		if (!do_listen) {
+			plog(LLV_DEBUG, LOCATION, NULL,
+				"ignore because do not listen on source address : %s.\n",
+				saddrwop2str(sa));
+			return 0;
+		}
+	}
 
 	/*
 	 * If there is a phase 2 handler against the policy identifier in
@@ -1777,8 +1861,7 @@ pk_recvdelete(mhp)
 		plog(LLV_DEBUG, LOCATION, NULL,
 			"%s message is not interesting "
 			"because the message was originated by me.\n",
-			s_pfkey_type(msg->sadb_msg_type),
-			msg->sadb_msg_pid);
+			s_pfkey_type(msg->sadb_msg_type));
 		return -1;
 	}
 
@@ -1878,6 +1961,9 @@ getsadbpolicy(policy0, policylen0, type, iph2)
 	xpl->sadb_x_policy_type = IPSEC_POLICY_IPSEC;
 	xpl->sadb_x_policy_dir = spidx->dir;
 	xpl->sadb_x_policy_id = 0;
+#ifdef HAVE_PFKEY_POLICY_PRIORITY
+	xpl->sadb_x_policy_priority = PRIORITY_DEFAULT;
+#endif
 
 	/* no need to append policy information any more if type is SPDDELETE */
 	if (type == SADB_X_SPDDELETE)
@@ -2005,6 +2091,16 @@ pk_recvspdupdate(mhp)
 	daddr = (struct sadb_address *)mhp[SADB_EXT_ADDRESS_DST];
 	xpl = (struct sadb_x_policy *)mhp[SADB_X_EXT_POLICY];
 
+#ifdef HAVE_PFKEY_POLICY_PRIORITY
+	KEY_SETSECSPIDX(xpl->sadb_x_policy_dir,
+			saddr + 1,
+			daddr + 1,
+			saddr->sadb_address_prefixlen,
+			daddr->sadb_address_prefixlen,
+			saddr->sadb_address_proto,
+			xpl->sadb_x_policy_priority,
+			&spidx);
+#else
 	KEY_SETSECSPIDX(xpl->sadb_x_policy_dir,
 			saddr + 1,
 			daddr + 1,
@@ -2012,6 +2108,7 @@ pk_recvspdupdate(mhp)
 			daddr->sadb_address_prefixlen,
 			saddr->sadb_address_proto,
 			&spidx);
+#endif
 
 	sp = getsp(&spidx);
 	if (sp == NULL) {
@@ -2095,6 +2192,16 @@ pk_recvspdadd(mhp)
 	daddr = (struct sadb_address *)mhp[SADB_EXT_ADDRESS_DST];
 	xpl = (struct sadb_x_policy *)mhp[SADB_X_EXT_POLICY];
 
+#ifdef HAVE_PFKEY_POLICY_PRIORITY
+	KEY_SETSECSPIDX(xpl->sadb_x_policy_dir,
+			saddr + 1,
+			daddr + 1,
+			saddr->sadb_address_prefixlen,
+			daddr->sadb_address_prefixlen,
+			saddr->sadb_address_proto,
+			xpl->sadb_x_policy_priority,
+			&spidx);
+#else
 	KEY_SETSECSPIDX(xpl->sadb_x_policy_dir,
 			saddr + 1,
 			daddr + 1,
@@ -2102,6 +2209,7 @@ pk_recvspdadd(mhp)
 			daddr->sadb_address_prefixlen,
 			saddr->sadb_address_proto,
 			&spidx);
+#endif
 
 	sp = getsp(&spidx);
 	if (sp != NULL) {
@@ -2180,6 +2288,16 @@ pk_recvspddelete(mhp)
 	daddr = (struct sadb_address *)mhp[SADB_EXT_ADDRESS_DST];
 	xpl = (struct sadb_x_policy *)mhp[SADB_X_EXT_POLICY];
 
+#ifdef HAVE_PFKEY_POLICY_PRIORITY
+	KEY_SETSECSPIDX(xpl->sadb_x_policy_dir,
+			saddr + 1,
+			daddr + 1,
+			saddr->sadb_address_prefixlen,
+			daddr->sadb_address_prefixlen,
+			saddr->sadb_address_proto,
+			xpl->sadb_x_policy_priority,
+			&spidx);
+#else
 	KEY_SETSECSPIDX(xpl->sadb_x_policy_dir,
 			saddr + 1,
 			daddr + 1,
@@ -2187,6 +2305,7 @@ pk_recvspddelete(mhp)
 			daddr->sadb_address_prefixlen,
 			saddr->sadb_address_proto,
 			&spidx);
+#endif
 
 	sp = getsp(&spidx);
 	if (sp == NULL) {
@@ -2224,6 +2343,16 @@ pk_recvspdexpire(mhp)
 	daddr = (struct sadb_address *)mhp[SADB_EXT_ADDRESS_DST];
 	xpl = (struct sadb_x_policy *)mhp[SADB_X_EXT_POLICY];
 
+#ifdef HAVE_PFKEY_POLICY_PRIORITY
+	KEY_SETSECSPIDX(xpl->sadb_x_policy_dir,
+			saddr + 1,
+			daddr + 1,
+			saddr->sadb_address_prefixlen,
+			daddr->sadb_address_prefixlen,
+			saddr->sadb_address_proto,
+			xpl->sadb_x_policy_priority,
+			&spidx);
+#else
 	KEY_SETSECSPIDX(xpl->sadb_x_policy_dir,
 			saddr + 1,
 			daddr + 1,
@@ -2231,6 +2360,7 @@ pk_recvspdexpire(mhp)
 			daddr->sadb_address_prefixlen,
 			saddr->sadb_address_proto,
 			&spidx);
+#endif
 
 	sp = getsp(&spidx);
 	if (sp == NULL) {
@@ -2288,6 +2418,16 @@ pk_recvspddump(mhp)
 		return -1;
 	}
 
+#ifdef HAVE_PFKEY_POLICY_PRIORITY
+	KEY_SETSECSPIDX(xpl->sadb_x_policy_dir,
+			saddr + 1,
+			daddr + 1,
+			saddr->sadb_address_prefixlen,
+			daddr->sadb_address_prefixlen,
+			saddr->sadb_address_proto,
+			xpl->sadb_x_policy_priority,
+			&spidx);
+#else
 	KEY_SETSECSPIDX(xpl->sadb_x_policy_dir,
 			saddr + 1,
 			daddr + 1,
@@ -2295,6 +2435,7 @@ pk_recvspddump(mhp)
 			daddr->sadb_address_prefixlen,
 			saddr->sadb_address_proto,
 			&spidx);
+#endif
 
 	sp = getsp(&spidx);
 	if (sp != NULL) {
@@ -2465,7 +2606,7 @@ pk_recv(so, lenp)
 u_int32_t
 pk_getseq()
 {
-	return arc4random();
+	return eay_random();
 }
 
 static int
@@ -2489,6 +2630,14 @@ addnewsp(mhp)
 	daddr = (struct sadb_address *)mhp[SADB_EXT_ADDRESS_DST];
 	xpl = (struct sadb_x_policy *)mhp[SADB_X_EXT_POLICY];
 
+#ifdef __linux__
+	/* bsd skips over per-socket policies because there will be no
+	 * src and dst extensions in spddump messages. On Linux the only
+	 * way to achieve the same is check for policy id.
+	 */
+	if (xpl->sadb_x_policy_id % 8 >= 3) return 0;
+#endif
+
 	new = newsp();
 	if (new == NULL) {
 		plog(LLV_ERROR, LOCATION, NULL,
@@ -2626,6 +2775,16 @@ addnewsp(mhp)
 		return -1;
 	}
 
+#ifdef HAVE_PFKEY_POLICY_PRIORITY
+	KEY_SETSECSPIDX(xpl->sadb_x_policy_dir,
+			saddr + 1,
+			daddr + 1,
+			saddr->sadb_address_prefixlen,
+			daddr->sadb_address_prefixlen,
+			saddr->sadb_address_proto,
+			xpl->sadb_x_policy_priority,
+			&new->spidx);
+#else
 	KEY_SETSECSPIDX(xpl->sadb_x_policy_dir,
 			saddr + 1,
 			daddr + 1,
@@ -2633,6 +2792,7 @@ addnewsp(mhp)
 			daddr->sadb_address_prefixlen,
 			saddr->sadb_address_proto,
 			&new->spidx);
+#endif
 
 	inssp(new);
 
