@@ -1,4 +1,4 @@
-/*	$KAME: sockmisc.c,v 1.40 2003/11/11 16:08:03 sakane Exp $	*/
+/* $Id: sockmisc.c,v 1.17 2004/11/20 16:16:59 monas Exp $ */
 
 /*
  * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
@@ -37,15 +37,16 @@
 #include <sys/uio.h>
 
 #include <netinet/in.h>
-#ifdef IPV6_INRIA_VERSION
+#ifndef HAVE_NETINET6_IPSEC
 #include <netinet/ipsec.h>
-#if defined(IP_RECVDSTADDR) && !defined(IPV6_RECVDSTADDR)
-#define IPV6_RECVDSTADDR IP_RECVDSTADDR
-#endif
 #else
 #include <netinet6/ipsec.h>
 #endif
 
+#if defined(IP_RECVDSTADDR) && !defined(IPV6_RECVDSTADDR)
+#define IPV6_RECVDSTADDR IP_RECVDSTADDR
+#endif
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -80,8 +81,8 @@ const int niflags = 0;
  */
 int
 cmpsaddrwop(addr1, addr2)
-	struct sockaddr *addr1;
-	struct sockaddr *addr2;
+	const struct sockaddr *addr1;
+	const struct sockaddr *addr2;
 {
 	caddr_t sa1, sa2;
 
@@ -133,8 +134,8 @@ cmpsaddrwop(addr1, addr2)
  */
 int
 cmpsaddrwild(addr1, addr2)
-	struct sockaddr *addr1;
-	struct sockaddr *addr2;
+	const struct sockaddr *addr1;
+	const struct sockaddr *addr2;
 {
 	caddr_t sa1, sa2;
 	u_short port1, port2;
@@ -198,8 +199,8 @@ cmpsaddrwild(addr1, addr2)
  */
 int
 cmpsaddrstrict(addr1, addr2)
-	struct sockaddr *addr1;
-	struct sockaddr *addr2;
+	const struct sockaddr *addr1;
+	const struct sockaddr *addr2;
 {
 	caddr_t sa1, sa2;
 	u_short port1, port2;
@@ -322,9 +323,9 @@ recvfromto(s, buf, buflen, flags, from, fromlen, to, tolen)
 	struct cmsghdr *cm;
 	struct iovec iov[2];
 	u_char cmsgbuf[256];
-#if defined(INET6) && defined(ADVAPI)
+#if defined(INET6) && defined(INET6_ADVAPI)
 	struct in6_pktinfo *pi;
-#endif /*ADVAPI*/
+#endif /*INET6_ADVAPI*/
 	struct sockaddr_in *sin;
 #ifdef INET6
 	struct sockaddr_in6 *sin6;
@@ -363,7 +364,7 @@ recvfromto(s, buf, buflen, flags, from, fromlen, to, tolen)
 		plog(LLV_ERROR, LOCATION, NULL,
 			"cmsg %d %d\n", cm->cmsg_level, cm->cmsg_type);)
 #endif
-#if defined(INET6) && defined(ADVAPI)
+#if defined(INET6) && defined(INET6_ADVAPI)
 		if (ss.ss_family == AF_INET6
 		 && cm->cmsg_level == IPPROTO_IPV6
 		 && cm->cmsg_type == IPV6_PKTINFO
@@ -487,7 +488,8 @@ sendfromto(s, buf, buflen, src, dst, cnt)
 	}
 
 	switch (src->sa_family) {
-#if defined(INET6) && defined(ADVAPI) && !defined(IPV6_INRIA_VERSION)
+#if defined(INET6) && defined(INET6_ADVAPI)
+// XXX: This block wasn't compiled on Linux - does it work?
 	case AF_INET6:
 	    {
 		struct msghdr m;
@@ -559,6 +561,7 @@ sendfromto(s, buf, buflen, src, dst, cnt)
 		return len;
 	    }
 #endif
+#ifdef __linux__
 	case AF_INET:
 	    {
 		struct msghdr m;
@@ -589,7 +592,7 @@ sendfromto(s, buf, buflen, src, dst, cnt)
 		cm->cmsg_level = IPPROTO_IP;
 		cm->cmsg_type = IP_PKTINFO;
 		pi = (struct in_pktinfo *)CMSG_DATA(cm);
-		memcpy(&pi->ipi_addr, &src6.sin_addr, sizeof(src6.sin_addr));
+		memcpy(&pi->ipi_spec_dst, &src6.sin_addr, sizeof(src6.sin_addr));
 		pi->ipi_ifindex = ifindex;
 
 		plog(LLV_DEBUG, LOCATION, NULL,
@@ -615,6 +618,7 @@ sendfromto(s, buf, buflen, src, dst, cnt)
 
 		return len;
 	    }
+#endif /* __linux__ */
 	default:
 	    {
 		int needclose = 0;
@@ -814,15 +818,19 @@ saddr2str(saddr)
 	if (saddr == NULL)
 		return NULL;
 
-	GETNAMEINFO(saddr, addr, port);
-	snprintf(buf, sizeof(buf), "%s[%s]", addr, port);
+	if (saddr->sa_family == AF_UNSPEC)
+		snprintf (buf, sizeof(buf), "%s", "anonymous");
+	else {
+		GETNAMEINFO(saddr, addr, port);
+		snprintf(buf, sizeof(buf), "%s[%s]", addr, port);
+	}
 
 	return buf;
 }
 
 char *
 saddrwop2str(saddr)
-	struct sockaddr *saddr;
+	const struct sockaddr *saddr;
 {
 	static char buf[NI_MAXHOST + NI_MAXSERV + 10];
 	char addr[NI_MAXHOST];
@@ -830,12 +838,48 @@ saddrwop2str(saddr)
 	if (saddr == NULL)
 		return NULL;
 
-	GETNAMEINFO(saddr, addr, NULL);
+	GETNAMEINFO_NULL(saddr, addr);
 	snprintf(buf, sizeof(buf), "%s", addr);
 
 	return buf;
 }
 
+char *
+naddrwop2str(const struct netaddr *naddr)
+{
+	static char buf[NI_MAXHOST + 10];
+	static const struct sockaddr sa_any;	/* this is initialized to all zeros */
+	
+	if (naddr == NULL)
+		return NULL;
+
+	if (memcmp(&naddr->sa, &sa_any, sizeof(sa_any)) == 0)
+		snprintf(buf, sizeof(buf), "%s", "any");
+	else {
+		snprintf(buf, sizeof(buf), "%s", saddrwop2str(&naddr->sa.sa));
+		snprintf(&buf[strlen(buf)], sizeof(buf) - strlen(buf), "/%ld", naddr->prefix);
+	}
+	return buf;
+}
+
+char *
+naddrwop2str_fromto(const char *format, const struct netaddr *saddr,
+		    const struct netaddr *daddr)
+{
+	static char buf[2*(NI_MAXHOST + NI_MAXSERV + 10) + 100];
+	char *src, *dst;
+
+	src = strdup(naddrwop2str(saddr));
+	dst = strdup(naddrwop2str(daddr));
+	/* WARNING: Be careful about the format string! Don't 
+	   ever pass in something that a user can modify!!! */
+	snprintf (buf, sizeof(buf), format, src, dst);
+	racoon_free (src);
+	racoon_free (dst);
+
+	return buf;
+}
+
 char *
 saddr2str_fromto(format, saddr, daddr)
 	const char *format;
@@ -925,7 +969,7 @@ mask_sockaddr(a, b, l)
 
 	if ((alen << 3) < l) {
 		plog(LLV_ERROR, LOCATION, NULL,
-			"unexpected inconsistency: %d %d\n", b->sa_family, l);
+			"unexpected inconsistency: %d %zu\n", b->sa_family, l);
 		exit(1);
 	}
 
@@ -935,9 +979,73 @@ mask_sockaddr(a, b, l)
 		p[i] = 0x00;
 }
 
+/* Compute a score describing how "accurate" a netaddr is for a given sockaddr.
+ * Examples:
+ * 	Return values for address 10.20.30.40 [port 500] and given netaddresses...
+ * 		10.10.0.0/16	=> -1	... doesn't match
+ * 		0.0.0.0/0	=> 0	... matches, but only 0 bits.
+ * 		10.20.0.0/16	=> 16	... 16 bits match
+ * 		10.20.30.0/24	=> 24	... guess what ;-)
+ * 		10.20.30.40/32	=> 32	... whole address match
+ * 		10.20.30.40:500	=> 33	... both address and port match
+ * 		10.20.30.40:501	=> -1	... port doesn't match and isn't 0 (=any)
+ */
+int
+naddr_score(const struct netaddr *naddr, const struct sockaddr *saddr)
+{
+	static const struct netaddr naddr_any;	/* initialized to all-zeros */
+	struct sockaddr sa;
+	uint16_t naddr_port, saddr_port;
+	int port_score;
+
+	if (!naddr || !saddr) {
+		plog(LLV_ERROR, LOCATION, NULL,
+		     "Call with null args: naddr=%p, saddr=%p\n",
+		     naddr, saddr);
+		return -1;
+	}
+
+	/* Wildcard address matches, but only 0 bits. */
+	if (memcmp(naddr, &naddr_any, sizeof(naddr_any)) == 0)
+		return 0;
+
+	/* If families don't match we really can't do much... */
+	if (naddr->sa.sa.sa_family != saddr->sa_family)
+		return -1;
+	
+	/* If port check fail don't bother to check addresses. */
+	naddr_port = extract_port(&naddr->sa.sa);
+	saddr_port = extract_port(saddr);
+	if (naddr_port == 0 || saddr_port == 0)	/* wildcard match */
+		port_score = 0;
+	else if (naddr_port == saddr_port)	/* exact match */
+		port_score = 1;
+	else					/* mismatch :-) */
+		return -1;
+
+	/* Here it comes - compare network addresses. */
+	mask_sockaddr(&sa, saddr, naddr->prefix);
+	if (loglevel >= LLV_DEBUG) {	/* debug only */
+		char *a1, *a2, *a3;
+		a1 = strdup(naddrwop2str(naddr));
+		a2 = strdup(saddrwop2str(saddr));
+		a3 = strdup(saddrwop2str(&sa));
+		plog(LLV_DEBUG, LOCATION, NULL,
+		     "naddr=%s, saddr=%s (masked=%s)\n",
+		     a1, a2, a3);
+		free(a1);
+		free(a2);
+		free(a3);
+	}
+	if (cmpsaddrwop(&sa, &naddr->sa.sa) == 0)
+		return naddr->prefix + port_score;
+
+	return -1;
+}
+
 /* Some usefull functions for sockaddr port manipulations. */
 u_int16_t
-extract_port (struct sockaddr *addr)
+extract_port (const struct sockaddr *addr)
 {
   u_int16_t port = -1;
   
